
import datetime as dt
from collections import deque

from sqlalchemy import inspect, select, and_
from sqlalchemy.sql.functions import coalesce

from ..lib.data import AttrDict
from ..squeal.tables.statistic import StatisticName, StatisticJournal, StatisticJournalInteger, StatisticJournalFloat


class WaypointReader:

    def __init__(self, log, with_timespan=True):
        self._log = log
        self._with_timespan = with_timespan

    def read(self, s, ajournal, names, owner):

        sn = inspect(StatisticName).local_table
        sj = inspect(StatisticJournal).local_table
        sji = inspect(StatisticJournalInteger).local_table
        sjf = inspect(StatisticJournalFloat).local_table

        id_map = self._id_map(s, ajournal, names, owner)
        ids = list(id_map.keys())

        for timespan in ajournal.timespans:
            self._log.debug('%s' % timespan)
            waypoint = None
            stmt = select([sn.c.id, sj.c.time, coalesce(sjf.c.value, sji.c.value)]) \
                .select_from(sj.join(sn).outerjoin(sjf).outerjoin(sji)) \
                .where(and_(sj.c.source_id == ajournal.id,
                            sn.c.id.in_(ids),
                            sj.c.time >= timespan.start,
                            sj.c.time <= timespan.finish)) \
                .order_by(sj.c.time)
            self._log.debug(stmt)
            for id, time, value in s.connection().execute(stmt):
                if waypoint and waypoint['time'] != time:
                    yield waypoint
                    waypoint = None
                if not waypoint:
                    waypoint = AttrDict({'time': time})
                    if self._with_timespan:
                        waypoint['timespan'] = timespan
                waypoint[id_map[id]] = value
        self._log.debug('Waypoints generated')

    def _id_map(self, s, ajournal, names, owner):
        # need to convert from statistic_name_id to attribute name
        return dict((self._id(s, ajournal, key, owner), value) for key, value in names.items())

    def _id(self, s, ajournal, name, owner):
        return s.query(StatisticName.id). \
            filter(StatisticName.name == name,
                   StatisticName.owner == owner,
                   StatisticName.constraint == ajournal.activity_group).scalar()


class Chunk:
    '''
    A collection of data points in time order, associated with a single timespan.

    In most of the uses below the contents are slowly incremented over time (and
    values popped off the front) as various statistics are calculated.
    '''

    def __init__(self, waypoint):
        self.__timespan = waypoint.timespan
        self.__waypoints = deque([waypoint])

    def append(self, waypoint):
        self.__waypoints.append(waypoint)

    def popleft(self):
        return self.__waypoints.popleft()

    def __diff(self, index, attr, zero=0):
        if len(self.__waypoints) > 1:
            return attr(self.__waypoints[index]) - attr(self.__waypoints[0])
        else:
            return zero

    def distance(self):
        return self.__diff(-1, lambda w: w.distance)

    def distance_delta(self):
        return self.__diff(1, lambda w: w.distance)

    def time(self):
        return self.__diff(-1, lambda w: w.time, dt.timedelta(0)).total_seconds()

    def time_delta(self):
        return self.__diff(1, lambda w: w.time, dt.timedelta(0)).total_seconds()

    def heart_rates(self):
        return (waypoint.heart_rate for waypoint in self.__waypoints if waypoint.heart_rate is not None)

    def __len__(self):
        return len(self.__waypoints)

    def __getitem__(self, item):
        return self.__waypoints[item]

    def __bool__(self):
        return self.distance_delta() > 0


class Chunks:
    '''
    This returns a sequence of lists of chunks, with progressively more waypoints.  The chunks are
    the *same* list, but with more points added each time (ie mutated / extended),

    The caller should read chunks until they contain sufficient data (for whatever is being calculated)
    and then discard and call to maintain the same data length.
    '''

    def __init__(self, log, waypoints):
        self._log = log
        self._waypoints = waypoints

    def chunks(self):
        chunks, chunk_index = deque(), {}
        for waypoint in self._waypoints:
            timespan = waypoint.timespan
            if timespan in chunk_index:
                chunk_index[timespan].append(waypoint)
            else:
                chunk = Chunk(waypoint)
                chunk_index[timespan] = chunk
                chunks.append(chunk)
            yield chunks

    def complete(self):
        '''
        The final chunks, containing all waypoints.
        '''
        last = None
        for chunks in self.chunks():
            last = chunks
        return last

    @staticmethod
    def drop_first(chunks):
        chunks[0].popleft()
        while chunks and not chunks[0]:
            chunks.popleft()


def filter_none(names, waypoints):
    names = list(names)
    return [w for w in waypoints if all(n in w and w[n] is not None for n in names)]




import datetime as dt
from abc import abstractmethod

from .. import BasePipeline
from ...lib.date import local_date_to_time
from ...lib.date import to_date
from ...squeal import ActivityGroup, ActivityJournal
from ...squeal.tables.pipeline import Pipeline, PipelineType


def display_pipeline(log, session, factory, date, diary, schedule=None):
    '''
    schedule only sent for summary views.
    '''
    date = to_date(date)   # why is this needed?
    for cls, args, kargs in Pipeline.all(log, session, PipelineType.DIARY):
        log.info('Building %s (%s, %s)' % (cls, args, kargs))
        yield from cls(log, *args, diary=diary, **kargs).display(session, factory, date, schedule=schedule)


class Displayer(BasePipeline):

    def _on_init(self, *args, diary=None, **kargs):
        self._diary = diary
        super()._on_init(*args, **kargs)

    def display(self, s, f, date, schedule=None):
        if schedule:
            yield from self._display_schedule(s, f, date, schedule=schedule)
        else:
            yield from self._display_date(s, f, date)

    @abstractmethod
    def _display_schedule(self, s, f, date, schedule=None):
        raise NotImplementedError()

    @abstractmethod
    def _display_date(self, s, f, date):
        raise NotImplementedError()


class JournalDiary(Displayer):

    def _display_date(self, s, f, date):
        start = local_date_to_time(date)
        finish = start + dt.timedelta(days=1)
        for activity_group in s.query(ActivityGroup).order_by(ActivityGroup.sort).all():
            for journal in s.query(ActivityJournal). \
                    filter(ActivityJournal.finish >= start,
                           ActivityJournal.start < finish,
                           ActivityJournal.activity_group == activity_group). \
                    order_by(ActivityJournal.start).all():
                yield from self._journal_date(s, f, journal, date)

    @abstractmethod
    def _journal_date(self, s, f, ajournal, date):
        raise NotImplementedError()




from collections import namedtuple
from itertools import repeat, zip_longest

from .support import Named
from ...lib.data import WarnDict

TIMESTAMP_GLOBAL_TYPE = 253


class ScaledField(Named):

    def __init__(self, log, name, units, scale, offset, accumulate):
        super().__init__(log, name)
        self._units = units
        self._scale = 1 if scale is None else scale
        self._offset = 0 if offset is None else offset
        self._is_scaled = self._scale != 1 or self._offset != 0
        self._is_accumulate = accumulate

    def _parse_and_scale(self, type, data, count, endian, timestamp, accumulate, **options):
        values = type.parse(data, count, endian, timestamp, **options)
        if values is not None:
            if self._is_scaled:
                values = tuple(value / self._scale - self._offset for value in values)
            if self._is_accumulate:
                values = accumulate(self, values)
        yield self.name, (values, self._units)

    def post(self, message, types):
        """
        Called after message creation.
        """
        pass


class TypedField(ScaledField):

    def __init__(self, log, name, field_no, units, scale, offset, accumulate, field_type, types):
        super().__init__(log, name, units, scale, offset, accumulate)
        self.number = field_no
        if types.is_type(name):
            self.type = types.profile_to_type(name)
            log.warning('Overriding type (%s) for predefined %s' % (field_type, name))
        else:
            self.type = types.profile_to_type(field_type)

    def parse(self, data, count, endian, timestamp, references, accumulate, message, **options):
        yield from self._parse_and_scale(self.type, data, count, endian, timestamp, accumulate, **options)


class RowField(TypedField):

    def __init__(self, log, row, types):
        super().__init__(log, row.field_name, row.single_int(log, row.field_no), row.units,
                         row.single_int(log, row.scale), row.single_int(log, row.offset),
                         row.single_int(log, row.accumulate), row.field_type, types)


class DelegateField(ScaledField):

    def parse(self, data, count, endian, timestamp, references, accumulate, message, **options):
        # todo - do we need to worry about padding data?
        delegate = message.profile_to_field(self.name)
        if isinstance(delegate, RowField):
            yield from self._parse_and_scale(delegate.type, data, count, endian, timestamp, accumulate, **options)
        else:
            # on dangerous ground here.  docs are unclear.  we'll do a complete delegation
            # unless this is scaled, in which case we don't know how to both scale and
            # delegate
            if self._is_scaled:
                raise Exception('Scaled component is not a simple field')
            else:
                yield from delegate.parse(data, count, endian, timestamp, references, accumulate, message, **options)

    def size(self, message):
        delegate = message.profile_to_field(self.name)
        return delegate.type.size


class Zip:

    def _zip(self, *fields):
        return zip_longest(*(self.__split(field) for field in fields))

    def __split(self, field):
        if field:
            for value in str(field).split(','):
                yield value.strip()
        else:
            yield None


class CompositeField(Zip, TypedField):

    def __init__(self, log, row, types):
        super().__init__(log, row.field_name, row.single_int(log, row.field_no), row.units,
                         None, None, None, row.field_type, types)
        self.number = row.single_int(log, row.field_no)
        self.__components = []
        for (name, bits, units, scale, offset, accumulate) in \
                self._zip(row.components, row.bits, row.units, row.scale, row.offset, row.accumulate):
            self.__components.append((int(bits),
                                      DelegateField(log, name, units,
                                                    None if scale is None else float(scale),
                                                    None if offset is None else float(offset),
                                                    accumulate)))

    def post(self, message, types):
        pass

    def parse(self, data, count, endian, timestamp, references, accumulate, message, rtn_composite=False, **options):
        if rtn_composite:
            yield (self.name, (('COMPOSITE',), self._units))
        byteorder = ['little', 'big'][endian]
        bits = int.from_bytes(data, byteorder=byteorder)
        for nbits, field in self.__components:
            nbytes = max((nbits+7) // 8, field.size(message))
            data = (bits & ((1 << nbits) - 1)).to_bytes(nbytes, byteorder=byteorder)
            bits >>= nbits
            yield from field.parse(data, 1, endian, timestamp, references, accumulate, message,
                                   rtn_composite=rtn_composite, **options)


class DynamicField(Zip, RowField):

    def __init__(self, log, row, rows, types):
        super().__init__(log, row, types)
        self.__dynamic_lookup = WarnDict(log, 'No dynamic field for %r')
        self.references = []
        for row in rows.lookahead():
            if row and row.field_name and row.field_no is None:
                for name, value in self._zip(row.ref_name, row.ref_value):
                    # need to use a list (not set) to preserve order
                    # (and consistently disambiguate multiple matches)
                    if name not in self.references:
                        self.references.append(name)
                    self.__dynamic_lookup[(name, value)] = row.field_name
            else:
                break

    def post(self, message, types):
        # fill in values for when mapping is not used
        for (name, value), field in list(self.__dynamic_lookup.items()):
            value = types.profile_to_type(message.profile_to_field(name).type.name).profile_to_internal(value)
            self.__dynamic_lookup[(name, value)] = field

    def parse(self, data, count, endian, timestamp, references, accumulate, message, warn=False, **options):
        for name in self.references:
            if name in references:
                lookup = (name, references[name][0][0])  # drop units and take first value
                if lookup in self.__dynamic_lookup:
                    yield from message.profile_to_field(self.__dynamic_lookup[lookup]).parse(
                        data, count, endian, timestamp, references, accumulate, message, **options)
                    return
        if warn:
            self._log.warning('Could not resolve dynamic field %s' % self.name)
        yield from super().parse(data, count, endian, timestamp, references, accumulate, message, warn=warn, **options)


def MessageField(log, row, rows, types):
    log.debug('Parsing field %s' % row.field_name)
    if row.components:
        return CompositeField(log, row, types)
    else:
        peek = rows.peek()
        if peek and peek.field_name and peek.field_no is None:
            return DynamicField(log, row, rows, types)
        else:
            return RowField(log, row, types)


class Row(namedtuple('BaseRow',
                     'msg_name, field_no, field_name, field_type, array, components, scale, offset, ' +
                     'units, bits, accumulate, ref_name, ref_value, comment, products, example')):

    __slots__ = ()

    def __new__(cls, row):
        return super().__new__(cls, *tuple(row)[0:16])

    def single_int(self, log, value):
        try:
            return None if value is None else int(value)
        except ValueError:
            log.warning('Cannot parse "%s" as a single integer', value)
            return None




from collections import Counter
from itertools import chain

from . import WaypointCalculator
from .heart_rate import hr_zones_from_database
from ..names import ACTIVE_DISTANCE, MAX, M, ACTIVE_TIME, S, ACTIVE_SPEED, KMH, round_km, MEDIAN_KM_TIME, \
    PERCENT_IN_Z, PC, TIME_IN_Z, HR_MINUTES, MAX_MED_HR_M, BPM, MIN, CNT, SUM, AVG, MSR, summaries, HEART_RATE, DISTANCE
from ..waypoint import Chunks
from ...squeal import StatisticName


class ActivityStatistics(WaypointCalculator):

    def _filter_statistic_journals(self, q):
        return q.filter(StatisticName.name == ACTIVE_TIME)

    def _names(self):
        return {HEART_RATE: 'heart_rate',
                DISTANCE: 'distance'}

    def _add_stats_from_waypoints(self, s, ajournal, waypoints):
        totals = Totals(self._log, waypoints)
        self._add_float_stat(s, ajournal,  ACTIVE_DISTANCE, summaries(MAX, CNT, SUM, MSR), totals.distance, M)
        self._add_float_stat(s, ajournal, ACTIVE_TIME, summaries(MAX, SUM, MSR), totals.time, S)
        self._add_float_stat(s, ajournal, ACTIVE_SPEED, summaries(MAX, AVG, MSR), totals.distance * 3.6 / totals.time, KMH)
        for target in round_km():
            times = list(sorted(TimeForDistance(self._log, waypoints, target * 1000).times()))
            if not times:
                break
            median = len(times) // 2
            self._add_float_stat(s, ajournal, MEDIAN_KM_TIME % target, summaries(MIN, MSR), times[median], S)
        zones = hr_zones_from_database(self._log, s, ajournal.activity_group, ajournal.start)
        if zones:
            for (zone, frac) in Zones(self._log, waypoints, zones).zones:
                self._add_float_stat(s, ajournal, PERCENT_IN_Z % zone, None, 100 * frac, PC)
            for (zone, frac) in Zones(self._log, waypoints, zones).zones:
                self._add_float_stat(s, ajournal, TIME_IN_Z % zone, None, frac * totals.time, S)
            for target in HR_MINUTES:
                heart_rates = sorted(MedianHRForTime(self._log, waypoints, target * 60).heart_rates(), reverse=True)
                if heart_rates:
                    self._add_float_stat(s, ajournal, MAX_MED_HR_M % target, summaries(MAX, MSR), heart_rates[0], BPM)
        else:
            self._log.warning('No HR zones defined for %s or before' % ajournal.start)


class TimeForDistance(Chunks):

    def __init__(self, log, waypoints, distance):
        super().__init__(log, waypoints)
        self.__distance = distance

    def times(self):
        for chunks in self.chunks():
            distance = sum(chunk.distance() for chunk in chunks)
            if distance > self.__distance:
                while chunks and distance - chunks[0].distance_delta() > self.__distance:
                    distance -= chunks[0].distance_delta()
                    self.drop_first(chunks)
                time = sum(chunk.time() for chunk in chunks)
                yield time * self.__distance / distance


class MedianHRForTime(Chunks):

    def __init__(self, log, waypoints, time, max_gap=None):
        super().__init__(log, waypoints)
        self.__time = time
        self.__max_gap = 0.01 * time if max_gap is None else max_gap
        log.debug('Will reject gaps > %ds' % int(self.__max_gap))

    def _max_gap(self, chunks):
        return max(c1[0].timespan.start - c2[0].timespan.finish
                   for c1, c2 in zip(list(chunks)[1:], chunks)).total_seconds()

    def heart_rates(self):
        for chunks in self.chunks():
            while len(chunks) > 1 and self._max_gap(chunks) > self.__max_gap:
                self._log.debug('Rejecting chunk because of gap (%ds)' % int(self._max_gap(chunks)))
                chunks.popleft()
            time = sum(chunk.time() for chunk in chunks)
            if time > self.__time:
                while chunks and time - chunks[0].time_delta() > self.__time:
                    time -= chunks[0].time_delta()
                    self.drop_first(chunks)
                heart_rates = list(sorted(chain(*(chunk.heart_rates() for chunk in chunks))))
                if heart_rates:
                    median = len(heart_rates) // 2
                    yield heart_rates[median]


class Totals(Chunks):

    def __init__(self, log, waypoints):
        super().__init__(log, waypoints)
        all_chunks = self.complete()
        self.distance = sum(chunk.distance() for chunk in all_chunks)
        self.time = sum(chunk.time() for chunk in all_chunks)


class Zones(Chunks):

    def __init__(self, log, waypoints, zones):
        super().__init__(log, waypoints)
        # this assumes record data are evenly distributed
        self.zones = []
        all_chunks = self.complete()
        counts = Counter()
        lower_limit = 0
        for zone, upper_limit in enumerate(zones):
            for chunk in all_chunks:
                for heart_rate in chunk.heart_rates():
                    if heart_rate is not None:
                        if lower_limit <= heart_rate < upper_limit:
                            counts[zone] += 1  # zero-based (incremented below)
            lower_limit = upper_limit
        total = sum(counts.values())
        if total:
            for zone in range(len(zones)):
                self.zones.append((zone + 1, counts[zone] / total))




from abc import ABC, abstractmethod
from enum import IntEnum


class MatchType(IntEnum):
    '''
    Control how MBRs are matched.
    '''

    EQUALS = 0  # request exactly equal to node
    CONTAINED = 1  # request contained in node
    CONTAINS = 2  # request contains node
    INTERSECTS = 3  # request and node intersect


class BaseTree(ABC):

    # nodes in the tree are
    #   (height, [entry, entry, entry...])
    #     entry is (mbr, content)
    #       content is node if height > 0
    #                  (points, value) if height = 0
    # note that nodes and entries are different

    # insertion/retrieval semantics are subtly different to a hash table:
    #   you can ask for matches that are exactly equal, overlap, contain or are contained by the box
    #   you can store multiple values at the same key
    #   you can store the same value multiple times at the same key
    #   both the above mean that querying is an iterator over matches
    # these issues above affect deletion:
    #   you will delete everything you would match
    #   you can specify a value on deletion, which provides an additional constraint
    #   but you could still delete multiple nodes
    #   (even with equality matching, (key, value) isn't guaranteed unique)
    # in other words, if you want unique (key, value) or unique key you must enforce it yourself.

    def __init__(self, items=None, *, max_entries=3, min_entries=None,
                 subtrees_flag=True, default_match=MatchType.EQUALS, default_border=0):
        '''
        Create an empty tree.

        `items` allows construction from an iterable of `(points, value)` pairs
        (as returned by `.items()`),

        `max_entries` is the maximum number of children a node can have.
        This data structure was originally designed for mass storage, where such a parameter is important.
        For in-memory use, a smaller value gives more branching, reduced comparisons, and better performance.

        `subtrees_flag` enables insertion of entire subtrees (rather than leaves) during internal balancing.
        It seems to be what is described in the original paper and gives a small speedup in some cases.
        '''
        if not min_entries:
            min_entries = max_entries // 2
        if min_entries > max_entries // 2:
            raise Exception('Min number of entries in a node is too high')
        if min_entries < 1:
            raise Exception('Min number of entries in a node is too low')
        self.__max_entries = max_entries
        self.__min_entries = min_entries
        self.__subtrees_flag = subtrees_flag
        self.__default_match = default_match
        self.__default_border = default_border
        self.__root = (0, [])
        self.__size = 0
        self.__hash = 966038070
        self.add_all(items)

    @property
    def global_mbr(self):
        if self.__size:
            x1, y1, x2, y2 = self._mbr_of_entries(*self.__root[1])
            x1, y1 = self._denormalize_point((x1, y1))
            x2, y2 = self._denormalize_point((x2, y2))
            return x1, y1, x2, y2
        else:
            return None

    @property
    def min_entries(self):
        return self.__min_entries

    @property
    def max_entries(self):
        return self.__max_entries

    @property
    def height(self):
        return self.__root[0]

    def size(self):
        return self.__size

    def _check_points(self, points):
        try:
            _ = points[0][0]
        except Exception:
            raise Exception('The `points` argument is a sequence of (x, y) points. ' +
                            'You may have entered a single (x, y) point.')

    def get(self, points, value=None, match=None, border=None):
        '''
        An iterator over values of nodes that match the MBR for the given points.

        The `match` describes the kind of matching done.

        If `value` is given then only nodes with that value are found.

        `border` is added to the MBR (eg to account for errors).
        '''
        self._check_points(points)
        match = self.__default_match if match is None else match
        border = self.__default_border if border is None else border
        points = self._normalize_points(points)
        mbr_request = self._mbr_of_points(points, border=border)
        content_request = (points, value)
        for points_entry, value_entry in self.__get_leaf_contents(self.__root, mbr_request, content_request, match):
            yield value_entry

    def get_items(self, points, value=None, match=None, border=None):
        '''
        An iterator over (MBR, value) of nodes that match the MBR for the given points.

        The `match` describes the kind of matching done.

        If `value` is given then only nodes with that value are found.

        `border` is added to the MBR (eg to account for errors).
        '''
        self._check_points(points)
        match = self.__default_match if match is None else match
        border = self.__default_border if border is None else border
        points = self._normalize_points(points)
        mbr_request = self._mbr_of_points(points, border=border)
        content_request = (points, value)
        for points_entry, value_entry in self.__get_leaf_contents(self.__root, mbr_request, content_request, match):
            yield self._denormalize_points(points_entry), value_entry

    def __get_leaf_contents(self, node, mbr_request, content_request, match):
        '''
        Internal get from node.
        '''
        height, entries = node
        for mbr_entry, content_entry in entries:
            if height:
                if self.__descend(mbr_request, mbr_entry, match):
                    yield from self.__get_leaf_contents(content_entry, mbr_request, content_request, match)
            elif self.__match(mbr_request, mbr_entry, content_request, content_entry, match):
                yield content_entry

    def __descend(self, mbr_request, mbr_entry, match):
        '''
        Descend in search?
        '''
        return (match in (MatchType.EQUALS, MatchType.CONTAINED) and self._contains(mbr_entry, mbr_request)) or \
               (match in (MatchType.CONTAINS, MatchType.INTERSECTS) and self._intersects(mbr_entry, mbr_request))

    def __match(self, mbr_request, mbr_entry, content_request, content_entry, match):
        '''
        Match search?
        '''
        points_node, value_node = content_entry
        points_request, value_request = content_request
        return (value_request is None or value_request == value_node) and (
                (match == MatchType.EQUALS and points_request == points_node) or
                (match == MatchType.CONTAINED and self._contains(mbr_entry, mbr_request)) or
                (match == MatchType.CONTAINS and self._contains(mbr_request, mbr_entry)) or
                (match == MatchType.INTERSECTS and self._intersects(mbr_request, mbr_entry)))

    def add(self, points, value, border=None):
        '''
        Add a value at the MBR of the given points.

        `border` is added to the MBR (eg to account for errors).
        '''
        self._check_points(points)
        border = self.__default_border if border is None else border
        points = self._normalize_points(points)
        mbr_addition = self._mbr_of_points(points, border=border)
        content = (points, value)
        self.__add_to_root(0, mbr_addition, content)
        self.__update_state(1, content)

    def add_all(self, items, border=None):
        '''
        Add a sequence of (point, value) pairs.

        `border` is added to the MBR (eg to account for errors).
        '''
        if items:
            for points, value in items:
                self.add(points, value, border=border)

    def __update_state(self, delta, content):
        '''
        Update size and hash.
        '''
        self.__size += delta
        self.__hash ^= hash(content)

    def __add_to_root(self, target, mbr_addition, content):
        '''
        Internal add at root (handling split has different logic to other nodes).

        `target` is the height for insertion.  This allows for subtrees to be re-inserted as a whole.
        '''
        split = self.__add_to_node(self.__root, target, mbr_addition, content)
        if split:
            height = split[0][1][0] + 1
            self.__root = (height, split)

    def __add_to_node(self, node, target, mbr_addition, content):
        '''
        Internal add at node.

        `target` is the height for insertion.  This allows for subtrees to be re-inserted as a whole.
        '''
        height, entries = node
        if height != target:
            i_best, mbr_best = self._best(entries, mbr_addition, height)
            child = entries[i_best][1]
            # optimistically set mbr on way down (since we have the mbr); will be removed if split spills data over
            entries[i_best] = (mbr_best, child)
            split = self.__add_to_node(child, target, mbr_addition, content)
            if split:
                del entries[i_best]
                entries.extend(split)
        else:
            entries.append((mbr_addition, content))
        # at this point parent mbr still correct (set on way down)
        if len(entries) > self.__max_entries:
            return self._split(height, entries)

    def _best(self, entries, mbr, height):
        '''
        Given a list of nodes, find the node whose area increases least when the given MBR is added.

        Not private because accessed by split mixins.
        '''
        i_best, mbr_best, area_best, delta_area_best, len_best = None, None, None, None, None
        for i_child, (mbr_entry, content_entry) in enumerate(entries):
            if i_best is None:
                i_best = i_child
                mbr_best = self._mbr_of_mbrs(mbr_entry, mbr)
                area_best = self._area_of_mbr(mbr_best)
                delta_area_best = self._area_of_mbr(mbr_best) - self._area_of_mbr(mbr_entry)
                len_best = len(content_entry[1]) if content_entry[0] else 0
            else:
                mbr_merge = self._mbr_of_mbrs(mbr_entry, mbr)
                area_merge = self._area_of_mbr(mbr_merge)
                delta_area_merge = area_merge - self._area_of_mbr(mbr_entry)
                if delta_area_merge < delta_area_best or \
                        (delta_area_merge == delta_area_best and area_merge < area_best) or \
                        (delta_area_merge == delta_area_best and area_merge == area_best and
                         height and len(entries) < len_best):
                    i_best, mbr_best, area_best, delta_area_best, len_best = \
                        i_child, mbr_merge, area_merge, delta_area_merge, \
                        len(content_entry[1]) if content_entry[0] else 0
        return i_best, mbr_best

    def delete(self, points, value=None, match=None, border=None):
        '''
        Remove entries that match the MBR of the given points and optional value.

        `border` is added to the MBR (eg to account for errors).
        '''
        self._check_points(points)
        match = self.__default_match if match is None else match
        border = self.__default_border if border is None else border
        points = self._normalize_points(points)
        mbr_deletion = self._mbr_of_points(points, border=border)
        content_deletion = (points, value)
        count = 0
        try:
            while True:
                self.__delete_one_from_root(mbr_deletion, content_deletion, match)
                count += 1
        except KeyError:
            return count

    def delete_one(self, points, value=None, match=None, border=None):
        '''
        Remove a single entry that match the MBR of the given points and optional value.

        Raises `KeyError` if no entry exists.

        `border` is added to the MBR (eg to account for errors).
        '''
        self._check_points(points)
        match = self.__default_match if match is None else match
        border = self.__default_border if border is None else border
        points = self._normalize_points(points)
        mbr_deletion = self._mbr_of_points(points, border=border)
        content_deletion = (points, value)
        self.__delete_one_from_root(mbr_deletion, content_deletion, match)

    def __delete_one_from_root(self, mbr_deletion, content_deletion, match):
        '''
        Internal deletion from root (handling deletion of a node has different logic to internal nodes).
        '''
        found = self.__delete_one_from_node(self.__root, mbr_deletion, content_deletion, match, 2)
        if found:
            delete, inserts, content_found = found
            if delete:
                self.__root = (0, [])
            self.__reinsert(inserts)
            if len(self.__root[1]) == 1 and self.height:  # single child, not leaf
                self.__root = self.__root[1][0][1]  # contents of first child
            self.__update_state(-1, content_found)
        else:
            points, value = content_deletion
            raise KeyError('Failed to delete %s%s' % (points, '' if value is None else ' (value %s)' % value))

    def __delete_one_from_node(self, node, mbr_deletion, content_deletion, match, local_min):
        '''
        Internal deletion from node.
        '''
        height, entries = node
        for i, (mbr_entry, contents_entry) in enumerate(entries):
            if height:
                if self.__descend(mbr_deletion, mbr_entry, match):
                    found = self.__delete_one_from_node(contents_entry, mbr_deletion, content_deletion, match,
                                                        self.__min_entries)
                    if found:
                        delete, inserts, content_found = found
                        if delete:
                            del entries[i]
                            if len(entries) < local_min:
                                inserts.extend((height, entry) for entry in entries)
                                return True, inserts, content_found
                        else:
                            # something under entries[i] has been deleted so recalculate mbr
                            _, (height_children, entries_children) = entries[i]
                            new_mbr = self._mbr_of_entries(*entries_children)
                            entries[i] = (new_mbr, (height_children, entries_children))
                        return False, inserts, content_found
            elif self.__match(mbr_deletion, mbr_entry, content_deletion, contents_entry, match):
                del entries[i]
                if len(entries) < local_min:
                    return True, [(height, entry) for entry in entries], contents_entry
                else:
                    return False, [], contents_entry

    def __leaves(self, node, canary=True):
        '''
        Iterator over the leaves in a node.
        '''
        if canary is True:
            canary = self.__hash
        height, entries = node
        for mbr_entry, content_entry in entries:
            if canary is not False and canary != self.__hash:
                raise RuntimeError('Tree was mutated while iterating over contents')
            if height:
                yield from self.__leaves(content_entry, canary=canary)
            else:
                yield mbr_entry, content_entry

    def __reinsert(self, inserts):
        '''
        Reinsert sub-trees (including leaves) removed during re-arrangement on deletion.

        We try to insert whole sub-trees, but this is not always possible (if the height has changed).
        '''
        max_height = self.height
        for (height_entry, (mbr_entry, content_entry)) in inserts:
            if height_entry and (height_entry > max_height or not self.__subtrees_flag):
                # can't (or won't) add the entire tree, so re-add the leaves
                for mbr_leaf, content_leaf in self.__leaves(content_entry, False):
                    self.__add_to_root(0, mbr_leaf, content_leaf)
            else:
                self.__add_to_root(height_entry, mbr_entry, content_entry)

    # standard container API

    def __len__(self):
        '''
        Expose number of entries though usual Python API.
        '''
        return self.__size

    def keys(self):
        '''
        All MBRs.
        '''
        for mbr, (points, value) in self.__leaves(self.__root):
            yield self._denormalize_points(points)

    def values(self):
        '''
        All values.
        '''
        for mbr, (points, value) in self.__leaves(self.__root):
            yield value

    def items(self):
        '''
        All (points, value) pairs.
        '''
        for mbr, (points, value) in self.__leaves(self.__root):
            yield (points, value)

    def __contains__(self, points):
        '''
        Equivalent to calling get() with the standard arguments and testing for result.
        '''
        try:
            next(self.get(points))
            return True
        except StopIteration:
            return False

    def __hash__(self):
        '''
        Hash based only on contents.
        '''
        return self.__hash

    def __eq__(self, other):
        '''
        Equality based only on contents.
        '''
        if other is self:
            return True
        # important to use hash(self) and not self.__hash because hash() truncates bits
        if not isinstance(other, BaseTree) or self.__size != len(other) or hash(self) != hash(other):
            return False
        # todo - is it better to have something slower but memory efficient?
        return sorted(list(self.items())) == sorted(list(other.items()))

    def __iter__(self):
        '''
        Iterable over keys.
        '''
        return self.keys()

    def __getitem__(self, points):
        '''
        Call get() with equality and no value..
        '''
        return self.get(points)

    def __setitem__(self, points, value):
        '''
        Add a value at the MBR associated with the given points.
        '''
        self.add(points, value)

    def __delitem__(self, points):
        '''
        Delete all entries that match the MBR associated with the given points.
        '''
        self.delete(points)

    def __str__(self):
        return '%s RTree (%s leaves, %d height, %d-%d entries)' % \
               (self.split_algorithm, len(self), self.height, self.min_entries, self.max_entries)

    # utilities for debugging

    def dump(self):
        '''
        An iterator over the tree structure.
        '''
        height, entries = self.__root
        yield height + 1, self._mbr_of_entries(*entries), None  # implied mbr on all nodes
        yield from self._dump_node(self.__root)

    def _dump_node(self, node):
        '''
        Internal iterator over the tree structure.
        '''
        height, entries = node
        for mbr, content in entries:
            if height:
                yield height, mbr, None
                yield from self._dump_node(content)
            else:
                yield height, mbr, content

    def print(self):
        '''
        Dump the tree structure to stdout for debugging.
        '''
        for height, mbr, content in self.dump():
            print('%2d %s %s %s' %
                  (height, ' ' * (10 - height), mbr, '' if content is None else content))

    def assert_consistent(self):
        '''
        Make some basic tests of consistency.
        '''
        if self.__size and not self.__root[1]:
            raise Exception('Emtpy root (and size %d)' % self.__size)
        size = self._assert_consistent(self.__root)
        if size != self.__size:
            raise Exception('Unexpected number of leaves (%d != %d)' % (size, self.__size))

    def _assert_consistent(self, node):
        '''
        Internal tests of consistency.
        '''
        if node:
            height, entries = node
            if len(entries) < self.__min_entries and node != self.__root:
                raise Exception('Too few children at height %d' % height)
            if len(entries) > self.__max_entries:
                raise Exception('Too many children at height %d' % height)
            if height:
                count = 0
                for mbr_entry, content_entry in entries:
                    height_entry, entries_entry = content_entry
                    mbr_check = self._mbr_of_entries(*entries_entry)
                    if mbr_check != mbr_entry:
                        raise Exception('Bad MBR at height %d %s / %s' % (height, mbr_check, mbr_entry))
                    count += self._assert_consistent(content_entry)
                return count
            else:
                return len(entries)
        else:
            return 0

    # allow different coordinate systems
    # nothing above should depend on the exact representation of point or mbr

    @property
    @abstractmethod
    def split_algorithm(self):
        raise NotImplementedError()

    def _normalize_points(self, points):
        return tuple(self._normalize_point(p) for p in points)

    @abstractmethod
    def _normalize_point(self, point):
        raise NotImplementedError()

    def _denormalize_points(self, points):
        return tuple(self._denormalize_point(p) for p in points)

    def _denormalize_point(self, point):
        return point

    @abstractmethod
    def _mbr_of_points(self, points, border=0):
        raise NotImplementedError()

    def _mbr_of_entries(self, *entries):
        return self._mbr_of_mbrs(*(mbr for mbr, _ in entries))

    @abstractmethod
    def _mbr_of_mbrs(self, *mbrs):
        raise NotImplementedError()

    @abstractmethod
    def _intersects(self, mbr1, mbr2):
        raise NotImplementedError()

    @abstractmethod
    def _contains(self, outer, inner):
        raise NotImplementedError()

    @abstractmethod
    def _area_of_mbr(self, mbr):
        raise NotImplementedError()

    # allow different split algorithms

    @abstractmethod
    def _split(self, height, entries):
        raise NotImplementedError()

    @abstractmethod
    def _pick_seeds(self, entries):
        raise NotImplementedError()

    @abstractmethod
    def _pick_next(self, pair, entries):
        raise NotImplementedError()


class CartesianMixin:
    '''
    Basic support for (x,y) points..
    '''

    def _normalize_point(self, point):
        '''
        No normalizetion by default.
        '''
        return point

    def _mbr_of_points(self, points, border=0):
        '''
        Find the MBR of a set of points,
        '''
        xs, ys = zip(*points)
        return min(xs) - border, min(ys) - border, max(xs) + border, max(ys) + border

    def _mbr_of_mbrs(self, *mbrs):
        '''
        Find the MBR of a set of MBRs.
        '''
        x1s, y1s, x2s, y2s = zip(*mbrs)
        return min(x1s), min(y1s), max(x2s), max(y2s)

    def _intersects(self, mbr1, mbr2):
        '''
        Do the two MBR's intersect?
        '''
        x1, y1, x2, y2 = mbr1
        X1, Y1, X2, Y2 = mbr2
        return x1 <= X2 and x2 >= X1 and y1 <= Y2 and y2 >= Y1

    def _contains(self, outer, inner):
        '''
        Is the `inner` MBR contained by the `outer`?
        '''
        x1, y1, x2, y2 = outer
        X1, Y1, X2, Y2 = inner
        return x1 <= X1 and x2 >= X2 and y1 <= Y1 and y2 >= Y2

    def _area_of_mbr(self, mbr):
        '''
        Area of the MBR
        '''
        x1, y1, x2, y2 = mbr
        return (x2 - x1) * (y2 - y1)

    def __extremes(self, entries):
        '''
        Internal routine for linear seeds.

        Measure the lowest higher bound, left-most right bound, etc, from a collection of MBRs
        and records which MRB (index) is responsible for each.
        '''
        index, extreme = None, None
        for i, (mbr, _) in enumerate(entries):
            if not i:
                index = [0, 0, 0, 0]
                extreme = list(mbr)
            else:
                for j in range(2):
                    if mbr[j] > extreme[j]:   # want max of lower left
                        extreme[j] = mbr[j]
                        index[j] = i
                    if mbr[j+2] < extreme[j+2]:  # and min of upper right
                        extreme[j+2] = mbr[j+2]
                        index[j+2] = i
        return index, extreme

    def _pick_seeds(self, entries):
        '''
        Choose the two MBRs that are furthest apart.

        Only called from LinearSplitMixin.
        '''
        global_mbr = self.global_mbr
        norm_x, norm_y = global_mbr[2] - global_mbr[0], global_mbr[3] - global_mbr[1]
        norm = [norm_x, norm_y, norm_x, norm_y]
        index, extremes = self.__extremes(entries)
        extremes = [x / n for x, n in zip(extremes, norm)]
        if extremes[2] - extremes[0] > extremes[3] - extremes[1] and index[0] != index[2]:
            return index[0], index[2]
        elif extremes[2] - extremes[0] < extremes[3] - extremes[1] and index[1] != index[3]:
            return index[1], index[3]
        else:
            return 0, 1  # degenerate case :(

    def _exact_area_sum(self, mbr0, mbr1):
        '''
        Don't count the overlapping region twice.

        Only called from ExponentialSplitMixin.
        '''
        if not mbr0:
            return self._area_of_mbr(mbr1)
        elif not mbr1:
            return self._area_of_mbr(mbr0)
        else:
            outer = self._area_of_mbr(mbr0) + self._area_of_mbr(mbr1)
            x1, y1, x2, y2 = mbr0
            X1, Y1, X2, Y2 = mbr1
            xm1 = max(x1, X1)
            xm2 = min(x2, X2)
            ym1 = max(y1, Y1)
            ym2 = min(y2, Y2)
            if xm2 > xm1 and ym2 > ym1:
                inner = (xm2 - xm1) * (ym2 - ym1)
            else:
                inner = 0
            return outer - inner


class LatLonMixin(CartesianMixin):
    '''
    Normalize (lon, lat) so that "wrapping" in longitude is not possible unless it varies by 180 degrees.
    '''

    def __init__(self, *args, **kargs):
        self.__zero_lon = None
        super().__init__(*args, **kargs)

    def _normalize_angle(self, x):
        '''
        Reduce angle to (-180, 180]
        '''
        while x <= -180:
            x += 360
        while x > 180:
            x -= 360
        return x

    def _normalize_point(self, point):
        '''
        Normalize to a Cartesian patch centred on first point found.
        '''
        lon, lat = point
        if self.__zero_lon is None:
            self.__zero_lon = lon
        lon = self._normalize_angle(lon - self.__zero_lon)
        return lon, lat

    def _denormalize_point(self, point):
        '''
        Revert the normalization above.
        '''
        lon, lat = point
        lon = self._normalize_angle(lon + self.__zero_lon)
        return lon, lat


class LinearMixin:
    '''
    Simple node selection.
    '''

    @property
    def split_algorithm(self):
        return 'Linear'

    def _split(self, height, entries):
        '''
        Divide the entries into two,
        '''
        i, j = self._pick_seeds(entries)
        split = [(entries[i][0], (height, [entries[i]])), (entries[j][0], (height, [entries[j]]))]
        del entries[max(i, j)]
        del entries[min(i, j)]
        # indexing ugliness below is rebuilding tuple with new mbr
        while entries:
            if len(entries) < self.min_entries:
                for index in 0, 1:
                    split_entries = split[index][1][1]
                    if len(split_entries) + len(entries) == self.min_entries:
                        split_entries.extend(entries)
                        split[index] = (self._mbr_of_entries(*split_entries), split[index][1])
                        return split
            node, entries = self._pick_next(split, entries)
            index, mbr = self._best(split, node[0], height)
            split[index] = (mbr, split[index][1])
            split[index][1][1].append(node)
        return split

    # _pick_seeds is on CartesianMixin because it's not abstracted from the coord system.

    def _pick_next(self, pair, entries):
        entry = entries.pop()
        return entry, entries


class QuadraticMixin(LinearMixin):
    '''
    This does a better job of grouping nodes (than linear) and is not measurably slower.
    '''

    @property
    def split_algorithm(self):
        return 'Quadratic'

    def _pick_seeds(self, entries):
        n, area_delta_best, best = len(entries), None, None
        for i in range(n):
            mbr_i = entries[i][0]
            for j in range(i):
                mbr_j = entries[j][0]
                mbr_both = self._mbr_of_mbrs(mbr_i, mbr_j)
                area_delta = self._area_of_mbr(mbr_both) - self._area_of_mbr(mbr_i) - self._area_of_mbr(mbr_j)
                if area_delta_best is None or area_delta > area_delta_best:
                    area_delta_best, best = area_delta, (i, j)
        return best

    def _pick_next(self, pair, entries):
        area_delta_best, best = None, None
        for i in range(len(entries)):
            area_delta = abs((self._area_of_mbr(self._mbr_of_entries(entries[i], pair[0])) - self._area_of_mbr(pair[0][0])) -
                             (self._area_of_mbr(self._mbr_of_entries(entries[i], pair[1])) - self._area_of_mbr(pair[1][0])))
            if area_delta_best is None or area_delta > area_delta_best:
                area_delta_best, best = area_delta, i
        node = entries[best]
        del entries[best]
        return node, entries


class ExponentialMixin:
    '''
    Pick nodes by making a complete comparison of all possibilities.
    '''

    @property
    def split_algorithm(self):
        return 'Exponential'

    def _pick_next(self, pair, nodes):
        # avoid ABC error
        raise NotImplementedError()

    def _pick_seeds(self, nodes):
        # avoid ABC error
        raise NotImplementedError()

    # _exact_area_sum() is on CartesianMixin because it is not abstracted from the coordinate system.

    def __split_recursive(self, mbr0, mbr1, entries, path=(), best=(None, None)):
        '''
        On each call we place the next node in either branch.
        To reduce memory usage we track only the path (successive indices) and best (lowest) area.
        '''

        # prune if too many nodes on one side
        depth = len(path)
        if path:
            remain = len(entries) - depth
            size1 = sum(path)
            size0 = depth - size1
            if size0 + remain < self.min_entries or size1 + remain < self.min_entries:
                return best

        if depth == len(entries):
            # we're done, so assess any improvement
            area = self._exact_area_sum(mbr0, mbr1)
            if best[0] is None or area < best[0]:
                best = (area, path)
        else:
            mbr = entries[depth][0]
            best = self.__split_recursive(self._mbr_of_mbrs(mbr0, mbr) if mbr0 else mbr, mbr1, entries, (*path, 0), best)
            best = self.__split_recursive(mbr0, self._mbr_of_mbrs(mbr1, mbr) if mbr1 else mbr, entries, (*path, 1), best)
        return best

    def _split(self, height, entries):
        '''
        Divide the nodes into two after considering all combinations.
        '''
        _, path = self.__split_recursive(None, None, entries)
        mbrs_pair, entries_pair = [None, None], [[], []]
        for i_node, i_pair in enumerate(path):
            node = entries[i_node]
            mbr_pair = mbrs_pair[i_pair]
            mbr_node = node[0]
            if mbr_pair is None:
                mbr_pair = mbr_node
            else:
                mbr_pair = self._mbr_of_mbrs(mbr_pair, mbr_node)
            mbrs_pair[i_pair] = mbr_pair
            entries_pair[i_pair].append(node)
        return [(mbrs_pair[0], (height, entries_pair[0])), (mbrs_pair[1], (height, entries_pair[1]))]


class CLRTree(LinearMixin, CartesianMixin, BaseTree): pass


class CQRTree(QuadraticMixin, CartesianMixin, BaseTree): pass


class CERTree(ExponentialMixin, CartesianMixin, BaseTree): pass


class LLRTree(LinearMixin, LatLonMixin, BaseTree): pass


class LQRTree(QuadraticMixin, LatLonMixin, BaseTree): pass


class LERTree(ExponentialMixin, LatLonMixin, BaseTree): pass






from bokeh.server.server import Server


class SingleShotServer:

    def __init__(self, log, plot, template=None, title=None, template_vars=None):
        self.__log = log
        self.__plot = plot
        self.__template = template
        self.__title = title
        self.__template_vars = {} if template_vars is None else template_vars
        self.__server = Server(self.__modify_doc)

        self.__server.start()
        self.__log.info('Opening Bokeh application on http://localhost:5006/')
        self.__server.io_loop.add_callback(self.__server.show, "/")
        self.__server.io_loop.call_later(5, self.__stop)
        self.__server.io_loop.start()

    def __modify_doc(self, doc):
        doc.add_root(self.__plot)
        doc.title = self.__title
        # this extends bokeh.core.templates.FILE - see code in bokeh.embed.elements
        doc.template = self.__template
        doc.template_variables.update(self.__template_vars)

    def __stop(self):
        self.__log.info('Stopping server')
        self.__server.io_loop.stop()
        self.__server.stop()
        self.__server.unlisten()




from json import dumps

from .database import add_statistics, add_enum_constant, set_constant, name_constant
from ..squeal.types import short_cls
from ..stoats.calculate.heart_rate import HeartRateStatistics, HRImpulse
from ..stoats.calculate.impulse import Response, ImpulseStatistics
from ..stoats.names import HR_IMPULSE, FITNESS, FATIGUE
from ..stoats.read.segment import SegmentImporter

FITNESS_CNAME = 'Fitness'
FATIGUE_CNAME = 'Fatigue'


def add_impulse(s, c, activity_group):
    '''
    Add configuration for a fitness/fatigue impulse model based on HR zones.

    This adds:
    * HeartRateStatistics pipeline class to calculate HR zone and impulse
      * Impulse constant with parameters for impulse calculation
    * ImpulseStatistics pipeline class to calculate fitness and fatigue
      * Fitness constant with parameters for fitness calculation
      * Fatigue constant with parameters for fatigue calculation
    '''

    activity_group_constraint = str(activity_group)

    hr_impulse_name = name_constant('HRImpulse', activity_group)
    hr_impulse = add_enum_constant(s, hr_impulse_name, HRImpulse, single=True, constraint=activity_group_constraint,
                                   description='Data needed to calculate the FF-model impulse from HR zones - ' +
                                               'see HRImpulse enum')
    set_constant(s, hr_impulse, dumps({'dest_name': HR_IMPULSE, 'gamma': 1.0, 'zero': 2, 'max_secs': 60}))

    # 7 and 42 days as for training peaks
    # https://www.trainingpeaks.com/blog/the-science-of-the-performance-manager/

    fitness_name = name_constant(FITNESS_CNAME, activity_group)
    fitness = add_enum_constant(s, fitness_name, Response, single=True, constraint=activity_group_constraint,
                                description='Data needed to calculate the FF-model fitness - see Response enum')
    set_constant(s, fitness, dumps({'src_name': HR_IMPULSE, 'src_owner': short_cls(HeartRateStatistics),
                                    'dest_name': FITNESS, 'tau_days': 42, 'scale': 1, 'start': 0}))

    fatigue_name = name_constant(FATIGUE_CNAME, activity_group)
    fatigue = add_enum_constant(s, fatigue_name, Response, single=True, constraint=activity_group_constraint,
                                description='Data needed to calculate the FF-model fitness - see Response enum')
    set_constant(s, fatigue, dumps({'src_name': HR_IMPULSE, 'src_owner': short_cls(HeartRateStatistics),
                                    'dest_name': FATIGUE, 'tau_days': 7, 'scale': 5, 'start': 0}))

    add_statistics(s, HeartRateStatistics, c, owner=short_cls(SegmentImporter), impulse=hr_impulse_name)
    add_statistics(s, ImpulseStatistics, c, responses=(fitness_name, fatigue_name), impulse=hr_impulse_name)






import datetime as dt
from hashlib import md5
from json import dumps, loads
from pydoc import locate
from struct import unpack

from sqlalchemy import TypeDecorator, Integer, Float, Text

from ..lib.date import to_time, to_date
from ..lib.schedule import Schedule


class Date(TypeDecorator):

    impl = Integer

    def process_literal_param(self, date, dialect):
        if date is None:
            return date
        return to_date(date).toordinal()

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        if value < 1:  # bootstrap TopicJournal.date where we set values to 0
            return None
        else:
            return dt.date.fromordinal(value)


class Time(TypeDecorator):

    impl = Float

    def process_literal_param(self, time, dialect):
        if time is None:
            return time
        else:
            return to_time(time).timestamp()

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            return dt.datetime.fromtimestamp(value, dt.timezone.utc)


CLS_CACHE = {}


class Cls(TypeDecorator):

    impl = Text

    def process_literal_param(self, cls, dialect):
        return long_cls(cls)

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        return lookup_cls(value)


def long_cls(cls):
    if cls is None:
        return cls
    if not isinstance(cls, str) and not isinstance(cls, type):
        cls = type(cls)
    if isinstance(cls, type):
        cls = cls.__module__ + '.' + cls.__name__
    return cls


def lookup_cls(value):
    # https://stackoverflow.com/a/24815361
    if not value:
        return None
    if value not in CLS_CACHE:
        CLS_CACHE[value] = locate(value)
    if not CLS_CACHE[value]:
        raise Exception('Cannot find %s' % value)
    return CLS_CACHE[value]


class ShortCls(TypeDecorator):

    impl = Text

    def process_literal_param(self, cls, dialect):
        return short_cls(cls)

    process_bind_param = process_literal_param


def short_cls(cls):
    if cls is None:
        return cls
    if not isinstance(cls, str) and not isinstance(cls, type):
        cls = type(cls)
    if isinstance(cls, type):
        cls = cls.__name__
    return cls


class Str(TypeDecorator):

    impl = Text

    def process_literal_param(self, value, dialect):
        if value is None:
            return value
        if isinstance(value, int):
            raise Exception('Passing primary key instead of class?')
        return str(value)

    process_bind_param = process_literal_param


class Json(TypeDecorator):

    impl = Text

    def process_literal_param(self, value, dialect):
        return dumps(value)

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        return loads(value)


class Sched(TypeDecorator):

    impl = Text

    def process_literal_param(self, sched, dialect):
        if sched is None:
            return sched
        if not isinstance(sched, Schedule):
            sched = Schedule(sched)
        return str(sched)

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return Schedule(value)


class OpenSched(Sched):

    def process_literal_param(self, sched, dialect):
        if sched is None:
            return sched
        if not isinstance(sched, Schedule):
            sched = Schedule(sched)
        sched.start = None
        sched.finish = None
        return str(sched)

    process_bind_param = process_literal_param


class Sort(TypeDecorator):

    impl = Integer

    def process_literal_param(self, value, dialect):
        if callable(value):
            value = value()
        return value

    process_bind_param = process_literal_param




from collections import deque

from ch2.command.args import ADD_HEADER, mm, HEADER_SIZE, PROFILE_VERSION, PROTOCOL_VERSION
from .format.tokens import FileHeader, token_factory, Checksum, State
from .profile.profile import read_profile


def fix(log, data, add_header=False, drop=False, slices=None, warn=False, force=True, validate=True, profile_path=None,
        header_size=None, protocol_version=None, profile_version=None,
        min_sync_cnt=3, max_record_len=None, max_drop_cnt=1, max_back_cnt=3, max_fwd_len=200):

    header_size = set_default(log, add_header, HEADER_SIZE, header_size, 14)
    protocol_version = set_default(log, add_header, PROTOCOL_VERSION, protocol_version, 0x10)  # from FR35
    profile_version = set_default(log, add_header, PROFILE_VERSION, profile_version, 0x07de)  # from FR35

    slices = parse_slices(log, slices)
    types, messages = read_profile(log, warn=warn, profile_path=profile_path)
    data = bytearray(data)

    if add_header:
        data = prepend_header(log, data, header_size, protocol_version, profile_version)

    if drop:
        slices = advance(log, State(log, types, messages), data, warn=warn, force=force,
                         min_sync_cnt=min_sync_cnt, max_record_len=max_record_len, max_drop_cnt=max_drop_cnt,
                         max_back_cnt=max_back_cnt, max_fwd_len=max_fwd_len)
        log.info('Dropped data to find slices: %s' % format_slices(slices))

    if slices:
        data = apply_slices(log, data, slices)
        log.info('Have %d bytes after slicing' % len(data))
        data += b'\0\0'
        log.warning('Appended blank checksum')

    data = fix_header(log, data)
    data = fix_checksum(log, data, State(log, types, messages))
    data = fix_header(log, data)  # if length changed with checksum

    if validate:
        validate_data(log, data, State(log, types, messages), warn=warn, force=force)

    return data


def set_default(log, required, name, value, deflt):
    if required and value is None:
        log.warning('Setting %s to %d' % (mm(name), deflt))
        return deflt
    else:
        return value


def parse_slices(log, slices):
    if not slices:
        return None
    parsed = [parse_slice(slice) for slice in slices.split(',')]
    if parsed and parsed[-1].stop is None:
        parsed[-1] = slice(parsed[-1].start, -2)
    log.debug('Parsed "%s" as "%s"' % (slices, format_slices(parsed)))
    return parsed


def parse_slice(s):
    start, stop = (parse_offset(offset) for offset in s.split(':'))
    return slice(start, stop)


def parse_offset(offset):
    if offset:
        return int(offset)
    else:
        return None


def format_slices(slices):
    return ','.join(format_slice(slice) for slice in slices)


def format_slice(slice):
    return '%s:%s' % (format_offset(slice.start), format_offset(slice.stop))


def format_offset(offset):
    return '' if offset in (0, None) else str(offset)


def prepend_header(log, data, header_size, protocol_version, profile_version):
    log.warning('Prepending file header of length %d' % header_size)
    data = bytearray([0] * header_size) + data
    header = FileHeader(data)
    header.repair(data, log,
                  header_size=header_size, protocol_version=protocol_version, profile_version=profile_version)
    data[:len(header)] = header.data
    return data


def fix_header(log, data):
    try:
        header = FileHeader(data)
        header.repair(data, log)
        data[:len(header)] = header.data
        return data
    except Exception as e:
        log.error(e)
        raise Exception('Error fixing header - maybe try %s' % mm(ADD_HEADER))


def fix_checksum(log, data, state):
    try:
        # check that we have the correct length
        deque(offset_tokens(state.copy(), data, warn=False, force=False), maxlen=0)
        checksum = Checksum(data[-2:])
        checksum.repair(data, log)
        data[-2:] = checksum.data
        return data
    except Exception as e:
        log.error(e)
        raise Exception('Cannot parse data to fix checksum')


def apply_slices(log, data, slices):
    result = bytearray()
    for slice in slices:
        result += data[slice]
    dropped = len(data) - len(result)
    if dropped:
        log.warning('Slicing decreased length by %d bytes' % dropped)
    return result


def validate_data(log, data, state, warn=False, force=True):
    try:
        file_header = FileHeader(data)
        file_header.validate(data, log)
        offset = len(file_header)
        while len(data) - offset > 2:
            token = token_factory(data[offset:], state)
            if token.is_user:
                record = token.parse(warn=warn)
                if force:
                    record.force()
            offset += len(token)
        checksum = Checksum(data[offset:])
        checksum.validate(data, log)
        log.info('Validated')
    except Exception as e:
        log.error(e)
        log.info('Validation failed')
        raise


def offset_tokens(state, data, offset=0, warn=False, force=True):
    '''
    this yields the offsets *after* the tokens (unlike tokens() in the read module).
    '''
    if not offset:
        file_header = FileHeader(data)
        offset = len(file_header)
        yield offset, file_header
    while len(data) - offset > 2:
        token = token_factory(data[offset:], state)
        if token.is_user:
            record = token.parse(warn=warn)
            if force:
                record.force()
        offset += len(token)
        yield offset, token
    # if we're here then there are 2 or less bytes remaining
    if len(data) - offset != 2:
        raise Exception('Misaligned checksum')


def slurp(log, state, data, initial_offset, warn=False, force=True, max_record_len=None):
    '''
    read as much as possible starting at the given offset.

    return offsets of valid tokens and a flag indicating if all data were read.
    '''
    offsets_and_states = []
    offset = initial_offset
    try:
        for offset, token in offset_tokens(state, data, initial_offset, warn=warn, force=force):
            if max_record_len and len(token) > max_record_len:
                log.info('Record too large (%d > %d) at offset %d' % (len(token), max_record_len, offset))
                return offsets_and_states, False
            offsets_and_states.append((offset, state.copy()))
        log.info('Read complete from %d' % initial_offset)
        return offsets_and_states, True
    except Exception as e:
        log.debug('%s at %d' % (e, offset))
        log.debug('Reading from offset %s found %d tokens before %d' %
                  (initial_offset, len(offsets_and_states), offset))
        return offsets_and_states, False


class Backtrack(Exception): pass


def advance(log, initial_state, data, drop_count=0, initial_offset=0, warn=False, force=True,
            min_sync_cnt=3, max_record_len=None, max_drop_cnt=1, max_back_cnt=3, max_fwd_len=200):
    '''
    try to synchronize on the stream and then read as much as possible.  if reading later fails,
    then try recurse if not at limit; if that fails, backtrack a record and try again.

    finally return the offsets used.

    "synchronize" means progressively discard data until some number of records smaller than
    some length can be read.
    '''
    offsets_and_states, complete = slurp(log, initial_state, data, initial_offset,
                                         warn=warn, force=force, max_record_len=max_record_len)
    if offsets_and_states:
        log.debug('%d: Read %d records; offset %d to %d' %
                  (drop_count, len(offsets_and_states), initial_offset, offsets_and_states[-1][0]))
    else:
        log.debug('%d: Did not read any records' % drop_count)
    offsets_and_states = [(0, initial_state)] + offsets_and_states
    if complete:
        # use explicit offset rather than open interval because need to drop final checksum
        return [slice(initial_offset, offsets_and_states[-1][0])]
    if drop_count and len(offsets_and_states) < min_sync_cnt:  # failing to sync on first read is OK
        raise Backtrack('Failed to sync at offset %d' % initial_offset)
    if drop_count >= max_drop_cnt:
        raise Backtrack('No more drops')
    for back_cnt in range(1, min(max_back_cnt + 1, len(offsets_and_states))):
        offset, state = offsets_and_states[-back_cnt]
        log.info('Searching forwards from offset %d after dropping %d records' % (offset, back_cnt-1))
        for delta in range(1, max_fwd_len):
            if offset + delta >= len(data):  # > for when a delta of 0 would have done
                log.info('Exhausted data')
                return [slice(initial_offset, offset)]
            else:
                try:
                    log.debug('%d: Retrying (drop %d, skip %d) at offset %d' %
                              (drop_count, back_cnt-1, delta, offset+delta))
                    slices = advance(log, state.copy(), data, drop_count+1, offset+delta,
                                     min_sync_cnt=min_sync_cnt, max_record_len=max_record_len,
                                     max_drop_cnt=max_drop_cnt, max_fwd_len=max_fwd_len)
                    return [slice(initial_offset, offset)] + slices
                except Backtrack as e:
                    log.debug('%d: Backtrack at (drop %d, skip %d): "%s"' % (drop_count, back_cnt, delta, e))
    raise Backtrack('Search exhausted at %d' % initial_offset)





from .heart_rate import *
from .plot import col_to_boxstats, bokeh_boxplot, closed_patch
from .data_frame import df, session, get_log, statistics, statistic_quartiles
from ..stoats.display.nearby import nearby_earlier, nearby_any_time

# avoid cleaning of imports
nearby_earlier, nearby_any_time




from .database import Counter, add_statistics, add_activity_group, add_activity_constant, add_topic, add_topic_field, \
    add_diary, add_activities, add_monitor, name_constant, add_nearby
from .impulse import add_impulse, FITNESS_CNAME, FATIGUE_CNAME
from ..lib.schedule import Schedule
from ..squeal.tables.statistic import StatisticJournalType
from ..squeal.tables.topic import TopicJournal
from ..squeal.types import short_cls
from ..stoats.calculate.activity import ActivityStatistics
from ..stoats.calculate.monitor import MonitorStatistics
from ..stoats.calculate.segment import SegmentStatistics
from ..stoats.calculate.summary import SummaryStatistics
from ..stoats.display.activity import ActivityDiary
from ..stoats.display.impulse import ImpulseDiary
from ..stoats.display.monitor import MonitorDiary
from ..stoats.display.nearby import NearbyDiary
from ..stoats.display.segment import SegmentDiary
from ..stoats.names import BPM, FTHR, MAX_HR, MAX_SPEED
from ..stoats.read.monitor import MonitorImporter
from ..stoats.read.segment import SegmentImporter
from ..uweird.fields.topic import Text, Float, Score0


def default(log, db, no_diary=False):

    with db.session_context() as s:

        # the following users helper functions (add_...) but you can also
        # execute arbitrary python code, use the session, etc.

        # basic activities

        c = Counter()
        bike = add_activity_group(s, 'Bike', c, description='All cycling activities')
        run = add_activity_group(s, 'Run', c, description='All running activities')
        # sport_to_activity maps from the FIT sport field to the activity defined above
        add_activities(s, SegmentImporter, c, sport_to_activity={'cycling': bike.name,
                                                                  'running': run.name})

        # statistics pipeline (called to calculate missing statistics)

        c = Counter()
        # need to specify the owner so that we get load waypoints correctly
        add_statistics(s, ActivityStatistics, c, owner=short_cls(SegmentImporter))
        add_statistics(s, SegmentStatistics, c, owner=short_cls(SegmentImporter))
        add_statistics(s, MonitorStatistics, c)
        add_impulse(s, c, bike)  # parameters set here can be adjusted via constants command

        # need to call normalize here because schedule isn't a schedule type column,
        # but part of a kargs JSON blob.
        add_statistics(s, SummaryStatistics, c, schedule=Schedule.normalize('m'))
        add_statistics(s, SummaryStatistics, c, schedule=Schedule.normalize('y'))

        # obviously you need to edit these parameters - see `ch2 constants Nearby.Bike`
        add_nearby(s, c, bike, 'Santiago', -33.4, -70.4)

        # diary pipeline (called to display data in the diary)

        c = Counter()
        add_diary(s, MonitorDiary, c)
        # these tie-in to the constants used in add_impulse()
        add_diary(s, ImpulseDiary, c,
                  fitness=name_constant(FITNESS_CNAME, bike),
                  fatigue=name_constant(FATIGUE_CNAME, bike))
        add_diary(s, ActivityDiary, c)
        add_diary(s, SegmentDiary, c)
        add_diary(s, NearbyDiary, c)

        # monitor pipeline

        c = Counter()
        add_monitor(s, MonitorImporter, c)

        # constants used by statistics

        add_activity_constant(s, bike, FTHR,
                              description='Heart rate at functional threshold (cycling). See https://www.britishcycling.org.uk/knowledge/article/izn20140808-Understanding-Intensity-2--Heart-Rate-0',
                              units=BPM, statistic_journal_type=StatisticJournalType.INTEGER)
        add_activity_constant(s, run, FTHR,
                              description='Heart rate at functional threshold (running).',
                              units=BPM, statistic_journal_type=StatisticJournalType.INTEGER)

        if not no_diary:

            # a basic diary

            c = Counter()
            diary = add_topic(s, 'Diary', c)
            add_topic_field(s, diary, 'Notes', c,
                            display_cls=Text)
            # now provided via monitor
            # add_topic_field(s, diary, 'Rest HR', c,
            #                 units=BPM, summary='[avg]',
            #                 display_cls=Integer, lo=25, hi=75)
            add_topic_field(s, diary, 'Weight', c,
                            units='kg', summary='[avg],[msr]',
                            display_cls=Float, lo=50, hi=100, dp=1)
            add_topic_field(s, diary, 'Sleep', c,
                            units='h', summary='[avg]',
                            display_cls=Float, lo=0, hi=24, dp=1)
            add_topic_field(s, diary, 'Mood', c,
                            summary='[avg]',
                            display_cls=Score0)
            add_topic_field(s, diary, 'Nutrition', c,
                            summary='[cnt]',
                            display_cls=Text)
            add_topic_field(s, diary, 'Soreness', c,
                            summary='[cnt]',
                            display_cls=Text)
            add_topic_field(s, diary, 'Medication', c,
                            summary='[cnt]',
                            display_cls=Text)
            add_topic_field(s, diary, 'Weather', c,
                            summary='[cnt]',
                            display_cls=Text)
            add_topic_field(s, diary, 'Route', c,
                            summary='[cnt]',
                            display_cls=Text)

        # finally, set the TZ so that first use of teh diary doesn't wipe all our intervals
        TopicJournal.check_tz(log, s)



import datetime as dt

from sqlalchemy import Column, Text, Integer, ForeignKey, UniqueConstraint
from sqlalchemy.orm import relationship, backref

from .source import Source, SourceType
from ..support import Base
from ..types import Time, Sort
from ...lib.date import format_time, local_date_to_time


class ActivityGroup(Base):

    __tablename__ = 'activity_group'

    id = Column(Integer, primary_key=True)
    name = Column(Text, nullable=False, server_default='')
    description = Column(Text, nullable=False, server_default='')
    sort = Column(Sort, nullable=False, server_default='')

    def __str__(self):
        return 'ActivityGroup "%s"' % self.name


class ActivityJournal(Source):

    __tablename__ = 'activity_journal'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    activity_group_id = Column(Integer, ForeignKey('activity_group.id'), nullable=False)
    activity_group = relationship('ActivityGroup')
    name = Column(Text, unique=True)
    fit_file = Column(Text, nullable=False, unique=True)
    start = Column(Time, nullable=False, index=True)
    finish = Column(Time, nullable=False)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.ACTIVITY
    }

    def __str__(self):
        return 'ActivityJournal %s %s to %s' % (self.activity_group.name,
                                                format_time(self.start), format_time(self.finish))

    def time_range(self, s):
        return self.start, self.finish

    @classmethod
    def at_date(cls, s, date):
        day = local_date_to_time(date)
        return s.query(ActivityJournal).filter(ActivityJournal.start >= day,
                                               ActivityJournal.start < day + dt.timedelta(days=1)).all()


class ActivityTimespan(Base):

    __tablename__ = 'activity_timespan'

    id = Column(Integer, primary_key=True)
    activity_journal_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'),
                                 nullable=False, index=True)
    activity_journal = relationship('ActivityJournal',
                                    backref=backref('timespans', cascade='all, delete-orphan',
                                                    passive_deletes=True,
                                                    order_by='ActivityTimespan.start'))
    start = Column(Time, nullable=False)
    finish = Column(Time, nullable=False)
    UniqueConstraint(activity_journal_id, start)

    def __str__(self):
        return 'ActivityTimespan from %s - %s' % (format_time(self.start), format_time(self.finish))






from math import sqrt
from random import uniform, gauss, seed, randrange
from time import time

from ch2.arty.spherical import Global
from ch2.arty.tree import CLRTree, MatchType, CQRTree, CERTree, LQRTree


def known_points(tree):
    tree.add([(0, 0)], '0')
    tree.add([(0, 1)], 'y')
    tree.add([(1, 0)], 'x')
    tree.add([(1, 1)], '1')
    assert len(tree) == 4, len(tree)
    all = list(tree.get([(0, 0), (1, 1)], match=MatchType.CONTAINS))
    assert len(all) == 4, all
    some = list(tree.get([(0, 0), (0, 1)], match=MatchType.CONTAINS))
    assert len(some) == 2, some
    some = list(tree.get([(0, 0), (0, 1)], match=MatchType.CONTAINED))
    assert len(some) == 0, some
    some = list(tree.get([(0, 0), (0, 1)], match=MatchType.EQUALS))
    assert len(some) == 0, some
    assert len(tree) == 4, len(tree)
    tree.delete([(0, 0)])
    assert len(tree) == 3, len(tree)
    some = list(tree.get([(0, 0), (0, 1)], match=MatchType.CONTAINS))
    assert len(some) == 1, some


def test_known_points():
    known_points(CLRTree(max_entries=2))
    known_points(CLRTree())
    known_points(CQRTree(max_entries=2))
    known_points(CQRTree())
    known_points(CERTree(max_entries=2))
    known_points(CERTree())


def known_boxes(tree):
    tree.add([(-0.1, -0.1), (0.1, 0.1)], '0')
    tree.add([(0.9, 0.9), (1.1, 1.1)], '1')
    tree.add([(0, 0), (1, 1)], 'sq')
    tree.add([(2, 2), (3, 3)], 'x')
    some = list(tree.get([(0, 0), (1, 1)], match=MatchType.CONTAINED))
    assert len(some) == 1, some
    some = list(tree.get([(0, 0), (1, 1)], match=MatchType.CONTAINS))
    assert len(some) == 1, some
    some = list(tree.get([(0, 0), (2, 1)], match=MatchType.CONTAINED))
    assert len(some) == 0, some
    some = list(tree.get([(0, 0), (2, 1)], match=MatchType.CONTAINS))
    assert len(some) == 1, some
    some = list(tree.get([(0, 0), (1, 1)], match=MatchType.INTERSECTS))
    assert len(some) == 3, some


def test_known_boxes():
    known_boxes(CLRTree(max_entries=2))
    known_boxes(CLRTree())
    known_boxes(CQRTree(max_entries=2))
    known_boxes(CQRTree())
    known_points(CERTree(max_entries=2))
    known_points(CERTree())


def test_delete_points():

    for size in 2, 3, 4, 8:

        def new_tree():
            tree = CQRTree(max_entries=size)
            for i in range(4):
                for j in range(4):
                    tree.add([(i, j)], i+j)
            assert len(tree) == 16
            return tree

        tree = new_tree()
        tree.delete([(1, 2)])
        assert len(tree) == 15, len(tree)

        tree = new_tree()
        tree.delete([(1, 1), (2, 2)])
        assert len(tree) == 16, len(tree)  # default is EQUALS
        tree = new_tree()
        tree.delete([(1, 1), (2, 2)], match=MatchType.CONTAINS)
        assert len(tree) == 12, len(tree)
        tree = new_tree()
        tree.delete([(1, 1), (2, 2)], match=MatchType.CONTAINED)
        assert len(tree) == 16, len(tree)


def random_box(n, size):
    x = uniform(0, size)
    y = uniform(0, size)
    dx = gauss(0, size/sqrt(n))
    dy = gauss(0, size/sqrt(n))
    return [(x, y), (x + dx, y + dy)]


def gen_random(n, size=100):
    for i in range(n):
        yield 1, random_box(n, size=size)


def test_equals():
    for size in 2, 3, 4, 8:
        tree1 = CQRTree(max_entries=size)
        for i, box in gen_random(10):
            tree1.add(box, i)
        tree2 = CQRTree(max_entries=size)
        for k, v in tree1.items():
            tree2.add(k, v)
        assert tree1 == tree2


def test_underscores():

    for size in 2, 3, 4, 8:
        seed(size)
        tree1 = CQRTree(max_entries=size)
        for i, box in gen_random(10):
            tree1[box] = i
        tree2 = CQRTree(max_entries=size)
        for k, v in tree1.items():
            tree2.add(k, v)
        assert tree1 == tree2
        if size == 2:
            assert str(tree1) == 'Quadratic RTree (10 leaves, 4 height, 1-2 entries)', str(tree1)
        tree3 = CQRTree(tree1.items(), max_entries=size)
        assert tree1 == tree3

    tree = CQRTree()
    tree.add([(1, 1)], '1')
    tree.add([(2, 2)], '2')
    assert len(tree) == 2
    assert [(1, 1)] in tree
    assert not [(3, 3)] in tree
    assert '1' in list(tree.values())
    assert ((1, 1),) in list(tree.keys()), list(tree.keys())
    assert (((1, 1),), '1') in list(tree.items())
    del tree[[(1, 1)]]
    assert len(tree) == 1
    assert list(tree[[(2, 2)]])
    assert list(tree.get([(2, 2)]))
    assert list(tree.get([(2, 2)], value='2'))
    assert not list(tree.get([(2, 2)], value='1'))
    assert tree
    del tree[[(2, 2)]]
    assert not tree


def best_bug(tree):
    for i, (value, box) in enumerate(gen_random(100)):
        tree.add(box, value % 10)
        tree.assert_consistent()


def test_best_bug():
    seed(4)  # 1:46 2:17 3:56 4:8 5:10 6:16 7:58 8:8
    best_bug(CLRTree(max_entries=2))
    best_bug(CQRTree(max_entries=2))
    best_bug(CERTree(max_entries=2))


def stress(type, n_children, n_data, check=True):
    seed(1)
    tree = type(max_entries=n_children)
    data = list(gen_random(n_data))

    for value, box in data:
        tree.add(box, value)
        if check:
            tree.assert_consistent()

    for i in range(10):

        n_delete = randrange(int(1.1 * n_data))   # delete to empty 10% of time
        for j in range(n_delete):
            if data:
                index = randrange(len(data))
                value, box = data[index]
                del data[index]
                tree.delete_one(box, value=value)
                if check:
                    tree.assert_consistent()
            else:
                assert not tree, tree

        while len(data) < n_data:
            value, box = next(gen_random(1))
            data.append((value, box))
            tree.add(box, value)
            if check:
                tree.assert_consistent()

        for j in range(n_data // 4):
            for match in range(4):
                list(tree.get(random_box(10, 100), match=MatchType(match)))


# def test_stress():
#     for type in CLRTree, CQRTree, CERTree:
#         print('type %s' % type)
#         for n_children in 3, 4, 10:
#             print('n_children %d' % n_children)
#             for n_data in 1, 2, 3, 100:
#                 print('n_data %d' % n_data)
#                 stress(type, n_children, n_data)


def test_latlon():
    tree = LQRTree()
    for lon in -180, 180:
        tree.add([(lon, 0)], str(lon))
    for lon in -180, 180:
        found = list(tree.get([(lon, 0)]))
        assert len(found) == 2, found
    area = tree._area_of_mbr(tree._mbr_of_points(tree._normalize_points([(-179, -1), (179, 1)])))
    assert area == 4, area

    tree = LQRTree()
    tree.add([(180, 0)], None)
    assert ((180, 0),) in list(tree.keys()), list(tree.keys())


def test_docs():

    tree = CQRTree()
    square = ((0,0),(0,1),(1,1),(1,0))
    tree[square] = 'square'
    assert list(tree[square]) == ['square']
    assert square in tree
    diagonal = ((0,0),(1,1))
    assert list(tree[diagonal]) == []
    assert not diagonal in tree
    assert list(tree.keys()) == [((0,0),(0,1),(1,1),(1,0))]
    assert list(tree.values()) == ['square']
    assert list(tree.items()) == [(((0,0),(0,1),(1,1),(1,0)), 'square')]
    assert len(tree) == 1
    del tree[square]
    assert len(tree) == 0

    tree = CQRTree(default_match=MatchType.INTERSECTS)
    tree[square] = 'square'
    assert list(tree[diagonal]) == ['square']

    tree = CQRTree(default_match=MatchType.INTERSECTS)
    tree[square] = 'square'
    assert list(tree.get_items(diagonal)) == [(((0,0),(0,1),(1,1),(1,0)), 'square')]


def test_canary():
    tree = CQRTree()
    for i in range(5):
        tree.add([(i, i)], i)
    keys = tree.keys()
    next(keys)
    del tree[[(1, 1)]]
    try:
        next(keys)
        assert False, 'expected error'
    except RuntimeError as e:
        assert 'mutated' in str(e), e


def measure(tree, n_data, n_loops, dim=100):

    seed(1)
    data = list(gen_random(n_data, size=dim))

    start = time()
    for i in range(n_loops):
        for value, box in data:
            tree[box] = value
        assert len(tree) == len(data), len(tree)
        for _, box in data:
            list(tree[box])
        for value, box in data:
            del tree[box]
        assert len(tree) == 0

    return time() - start


def measure_sizes():
    for type in CLRTree, CQRTree, CERTree:
        print()
        for size in 2, 3, 4, 6, 8, 10, 16, 32, 64, 128:
            for subtrees in True, False:
                t = measure(type(max_entries=size, subtrees_flag=subtrees), 1000, 2)
                print('%s %d %s %s' % (type, size, subtrees, t))
                if t > 5 and size > 4:
                    print('abort')
                    return


def test_global():

    def test_point(x, y, z):
        t = Global()
        t.add([(x, y)], z)
        l = list(t.get_items([(x, y)]))
        assert len(l) == 9, l
        for p, q in l:
            assert -0.001 < p[0][0] - x < 0.001, (p, (x, y))
            assert -0.001 < p[0][1] - y < 0.001, (p, (x, y))
            assert q == z, (q, z)

    test_point(0.01, 0.01, 0)
    test_point(179.9, 0.01, 1)
    test_point(-179.9, 0.01, 2)
    test_point(0.01, 89.99, 0)
    test_point(179.9, 89.99, 1)
    test_point(-179.9, 89.99, 2)
    test_point(0.01, -89.99, 0)
    test_point(179.9, -89.99, 1)
    test_point(-179.9, -89.99, 2)


        # for profiling
# PYTHONPATH=. python -m cProfile -s tottime tests/test_arty.py
if __name__ == '__main__':
    stress(CQRTree, 4, 1000, check=False)




from collections import Counter, defaultdict
from re import compile
from sys import stdout

from .format.read import filtered_records, filtered_tokens
from .format.records import no_bad_values, fix_degrees, append_units, no_unknown_fields, unique_names, join_values, \
    to_hex, no_filter
from ..command.args import MESSAGES, RECORDS, FIELDS, CSV, \
    TABLES, GREP
from ..lib.io import terminal_width
from ..lib.utils import unique


def summarize(log, format, fit_path, all_fields=False, all_messages=False, after=0, limit=-1,
              records=None, warn=False, profile_path=None, grep=None, name=False, invert=False, match=1,
              no_header=False):
    if name and format != GREP:
        print()
        print(fit_path)
    if format == RECORDS:
        summarize_records(log, fit_path,
                          all_fields=all_fields, all_messages=all_messages,
                          after=after, limit=limit, records=records, warn=warn, no_header=no_header,
                          profile_path=profile_path)
    elif format == TABLES:
        summarize_tables(log, fit_path,
                         all_fields=all_fields, all_messages=all_messages,
                         after=after, limit=limit, records=records, warn=warn, no_header=no_header,
                         profile_path=profile_path)
    elif format == GREP:
        summarize_grep(log, fit_path, grep, name_file=name, match=match, invert=invert,
                       after=after, limit=limit, warn=warn, no_header=no_header,
                       profile_path=profile_path)
    elif format == CSV:
        summarize_csv(log, fit_path,
                      after=after, limit=limit, warn=warn, no_header=no_header,
                      profile_path=profile_path)
    elif format == MESSAGES:
        summarize_messages(log, fit_path,
                           after=after, limit=limit, warn=warn, no_header=no_header,
                           profile_path=profile_path)
    elif format == FIELDS:
        summarize_fields(log, fit_path,
                         after=after, limit=limit, warn=warn, no_header=no_header,
                         profile_path=profile_path)
    else:
        raise Exception('Bad format: %s' % format)


def summarize_messages(log, fit_path, after=0, limit=-1, warn=False, no_header=False, profile_path=None):
    data, types, messages, tokens = \
        filtered_tokens(log, fit_path, after=after, limit=limit, warn=warn, no_header=no_header,
                        profile_path=profile_path)
    for index, offset, token in tokens:
        print('%03d %05d %s' % (index, offset, token))


def summarize_fields(log, fit_path, after=0, limit=-1, warn=False, no_header=False, profile_path=None):
    data, types, messages, tokens = \
        filtered_tokens(log, fit_path, after=after, limit=limit, warn=warn, no_header=no_header,
                        profile_path=profile_path)
    for index, offset, token in tokens:
        print('%03d %05d %s' % (index, offset, token))
        for line in token.describe_fields(types):
            print('  %s' % line)


def summarize_records(log, fit_path, all_fields=False, all_messages=False, after=0, limit=-1, records=None,
                      warn=False, no_header=False, profile_path=None):
    data, types, messages, records = \
        filtered_records(log, fit_path, after=after, limit=limit, records=records, warn=warn,
                         no_header=no_header, profile_path=profile_path)
    records = list(records)
    width = terminal_width()
    print()
    pprint_as_dicts(records, all_fields, all_messages, width=width)


def summarize_tables(log, fit_path, all_fields=False, all_messages=False, after=0, limit=-1, records=None,
                     warn=False, no_header=False, profile_path=None):
    data, types, messages, records = \
        filtered_records(log, fit_path, after=after, limit=limit, records=records, warn=warn,
                         no_header=no_header, profile_path=profile_path)
    records = list(records)
    counts = Counter(record.identity for record in records)
    small, large = partition(records, counts)
    width = terminal_width()
    print()
    pprint_as_dicts(small, all_fields, all_messages, width=width)
    pprint_as_tuples(large, all_fields, all_messages, width=width)


class Done(Exception):
    pass


def summarize_grep(log, fit_path, grep, name_file=False, match=1, invert=False, after=0, limit=-1,
                   warn=False, no_header=False, profile_path=None):
    data, types, messages, records = \
        filtered_records(log, fit_path, warn=warn, no_header=no_header, profile_path=profile_path)
    matchers = [compile(pattern) for pattern in grep]
    counts = defaultdict(lambda: 0)
    first = True
    try:
        for n, record in enumerate(records):
            if n >= after:
                if n < after + limit or limit < 0:
                    record = record.as_dict(join_values, append_units, to_hex, fix_degrees, no_bad_values)
                    for name, value in sorted(record.data.items()):
                        target_1 = '%s:%s' % (record.name, name)
                        target_2 = '%s:%s=%s' % (record.name, name, value)
                        for matcher in matchers:
                            if matcher.match(target_2 if '=' in matcher.pattern else target_1):
                                counts[matcher] += 1
                                if counts[matcher] <= match or match < 0:
                                    if first:
                                        print()
                                        first = False
                                    print('%s:%s=%s' % (record.name, name, value))
                                # exit early if we've displayed/matched all we need to
                                if match > -1 and all(counts[m] >= max(1, match) for m in matchers):
                                    raise Done()
                else:
                    raise Done()
    except Done:
        pass
    if name_file:
        if (not all(counts[m] for m in matchers)) == invert:
            print(fit_path)


def summarize_csv(log, fit_path, after=0, limit=-1, warn=False, no_header=False, profile_path=None,
                  out=stdout):
    data, types, messages, tokens = \
        filtered_tokens(log, fit_path, after=after, limit=limit, warn=warn, no_header=no_header,
                        profile_path=profile_path)
    for index, offset, token in tokens:
        if hasattr(token, 'describe_csv'):
            print(','.join(str(component) for component in token.describe_csv()), file=out)


def partition(records, counts, threshold=3):
    small, large = [], []
    for record in records:
        if counts[record.identity] <= threshold:
            small.append(record)
        else:
            large.append(record)
    return small, large


def pprint_as_dicts(records, all_fields, all_messages, width=80):
    for record in records:
        if all_messages or record.is_known():
            record = record.as_dict(join_values, append_units, to_hex, fix_degrees,
                                    no_filter if all_fields else no_unknown_fields,
                                    no_bad_values)
            print(record.identity)
            pprint_with_tabs(('%s: %s' % (name, value) for name, value in sorted(record.data.items())),
                             width=width)
            print()


def sort_names(data):
    return sorted(list(data), key=lambda x: ' ' if x[0] == 'timestamp' else x[0])


def pprint_as_tuples(records, all_fields, all_messages, width=80):
    records = [record.force(sort_names, unique_names,
                            timestamp=([record.timestamp], 's'))
               for record in records]
    titles = [record.as_names(no_filter if all_fields else no_unknown_fields)
              for record in unique(records, key=lambda x: x.identity)
              if all_messages or record.is_known()]
    for title in titles:
        pprint_series(title,
                      [record.as_values(join_values, append_units, to_hex, fix_degrees,
                                        no_filter if all_fields else no_unknown_fields)
                       for record in records
                       if record.identity == title.identity],
                      width=width)


def measure_lengths(records, lengths=None, separator=',', keep_bad=False):
    for record in records:
        if lengths is None:
            lengths = [(len(datum) + (len(separator) if i+1 < len(record.data) else 0)) if datum else 0
                       for i, datum in enumerate(record.data)]
        else:
            for i, datum in enumerate(record.data):
                if datum and (lengths[i] or not keep_bad):
                    lengths[i] = max(lengths[i], len(datum) + (len(separator) if i+1 < len(record.data) else 0))
    return lengths


def pad_to_lengths(record, lengths, separator=',', bad='-'):
    for i, datum in enumerate(record.data):
        if lengths[i]:
            if datum:
                yield (datum + (separator if i+1 < len(lengths) else '')).ljust(lengths[i])
            else:
                yield bad.ljust(lengths[i])


def pprint_series(title, records, width=80):
    print(title.identity)
    lengths = measure_lengths([title], keep_bad=True, lengths=measure_lengths(records))
    pprint_with_tabs(pad_to_lengths(title, lengths), first_indent=2, indent=4, separator='', width=width)
    for record in records:
        pprint_with_tabs(pad_to_lengths(record, lengths), first_indent=2, indent=4, separator='', width=width)
    print()


def pprint_with_tabs(data, first_indent=None, indent=2, tab=4, width=80, min_space=2, separator=','):
    if first_indent is None:
        first_indent = indent
    min_space -= 1
    line = ' ' * first_indent
    first = True
    for datum in data:
        if not first:
            line += separator
        l_old = len(line)
        if l_old > indent:
            l_new = (1 + (l_old - indent + min_space) // tab) * tab + indent
            # equality here leaves an attractive space at end of line
            if l_new + len(datum) >= width:
                print(line)
                line = ' ' * indent + datum
            else:
                line += ' ' * (l_new - l_old) + datum
        else:
            line += datum
        first = False
    if len(line) > indent:
        print(line)




from logging import basicConfig, DEBUG, getLogger
from sys import stdout
from unittest import TestCase

from ch2.fit.fix import fix
from ch2.fit.format.tokens import FileHeader
from ch2.fit.profile.profile import read_fit


class TestFixFit(TestCase):

    def setUp(self):
        basicConfig(stream=stdout, level=DEBUG)
        self.log = getLogger()

    def test_null(self):
        good = read_fit(self.log, '/home/andrew/project/ch2/choochoo/data/test/personal/2018-08-27-rec.fit')
        same = fix(self.log, bytearray(good))
        self.assertTrue(good is not same)  # check making a copy
        self.assertEqual(good, same)

    def test_null_drop(self):
        good = read_fit(self.log, '/home/andrew/project/ch2/choochoo/data/test/personal/2018-08-27-rec.fit')
        same = fix(self.log, bytearray(good), drop=True)
        self.assertTrue(good is not same)  # check making a copy
        self.assertEqual(good, same)

    def test_null_slices(self):
        good = read_fit(self.log, '/home/andrew/project/ch2/choochoo/data/test/personal/2018-08-27-rec.fit')
        same = fix(self.log, bytearray(good), slices=':')
        self.assertTrue(good is not same)  # check making a copy
        self.assertEqual(good, same)

    def test_drop(self):
        bad = read_fit(self.log, '/home/andrew/project/ch2/choochoo/data/test/other/8CS90646.FIT')
        good = fix(self.log, bad, drop=True)
        self.assertTrue(len(good) < len(bad))

    def test_slices(self):
        bad = read_fit(self.log, '/home/andrew/project/ch2/choochoo/data/test/other/8CS90646.FIT')
        with self.assertRaisesRegexp(Exception, 'Cannot parse data'):
            fix(self.log, bad, slices=':1000')  # first 1k bytes only

    def test_no_last_byte(self):
        good = read_fit(self.log, '/home/andrew/project/ch2/choochoo/data/test/personal/2018-08-27-rec.fit')
        same = fix(self.log, bytearray(good), drop=True)
        self.assertEqual(same, good)
        fixed = fix(self.log, bytearray(good)[:-1], drop=True)
        self.assertEqual(fixed, good)
        fixed = fix(self.log, bytearray(good)[:-2], drop=True)
        self.assertEqual(fixed, good)

    def test_no_header(self):
        good = read_fit(self.log, '/home/andrew/project/ch2/choochoo/data/test/personal/2018-08-27-rec.fit')
        same = fix(self.log, bytearray(good), drop=True)
        self.assertTrue(good is not same)  # check making a copy
        self.assertEqual(same, good)
        header = FileHeader(good)
        with self.assertRaisesRegex(Exception, 'Cannot parse data'):
            fix(self.log, bytearray(good)[len(header):])
        fixed = fix(self.log, bytearray(good)[len(header):], add_header=True, drop=True)
        self.assertEqual(good, fixed)
        fixed = fix(self.log, bytearray(good), add_header=True, slices=':14,28:')
        self.assertEqual(good, fixed)




import datetime as dt
from abc import abstractmethod
from collections import defaultdict, Counter
from struct import unpack, pack

from ..profile.fields import TypedField, TIMESTAMP_GLOBAL_TYPE, DynamicField
from ..profile.types import timestamp_to_time, time_to_timestamp
from ...lib.data import WarnDict, tohex


class Identity:

    def __init__(self, name, counter):
        self.name = name
        counter[name] += 1
        self.count = counter[name]
        self.__counter = counter

    def __repr__(self):
        current = self.__counter[self.name]
        if current == 1:
            return self.name
        else:
            return '%s (defn %d/%d)' % (self.name, self.count, current)


class Token:

    __slots__ = ('tag', 'is_user', 'data')

    def __init__(self, tag, is_user, data):
        self.tag = tag
        self.is_user = is_user
        self.data = data

    def __str__(self):
        return '%s %s' % (self.tag, tohex(self.data))

    def __len__(self):
        return len(self.data)


class ValidateToken(Token):

    @abstractmethod
    def validate(self, data, log, quiet=False):
        raise NotImplementedError()

    def _error(self, msg, log, quiet):
        if quiet:
            log.warning(msg)
        else:
            raise Exception(msg)


FIT = b'.FIT'


class FileHeader(ValidateToken):

    def __init__(self, data):
        super().__init__('HDR', False, data[:data[0]])
        self.header_size = data[0]
        self.protocol_version = data[1]
        self.profile_version = unpack('<H', data[2:4])[0]
        self.data_size = unpack('<I', data[4:8])[0]
        self.data_type = b''.join(unpack('4c', data[8:12]))
        if len(self) > 13:
            self.checksum = unpack('<H', data[12:14])[0]
            self.has_checksum = self.checksum != 0
        else:
            self.has_checksum = False

    def validate(self, data, log, quiet=False, header_size=None, protocol_version=None, profile_version=None):
        if self.header_size < 12:
            self._error('Header size too small (%d)' % self.header_size, log, quiet)
        if self.header_size not in (12, 14):
            log.warning('Header size not standard (%d/12/14)' % self.header_size)
        if header_size is not None:
            if self.header_size != header_size:
                self._error('Header size incorrect (%d/%d)' % (self.header_size, header_size), log, quiet)
        if protocol_version is not None:
            if self.profile_version != protocol_version:
                self._error('Protocol version incorrect (%d%d)' % (self.protocol_version, protocol_version), log, quiet)
        if profile_version is not None:
            if self.profile_version != profile_version:
                self._error('Profile version incorrect (%d%d)' % (self.profile_version, profile_version), log, quiet)
        if len(data) != self.data_size + len(self) + 2:
            self._error('Data size incorrect (%d/%d+%d+2=%d)' % (len(data), self.data_size, len(self),
                                                                 self.data_size + len(self) + 2), log, quiet)
        if self.data_type != FIT:
            self._error('Data type incorrect (%s)' % (self.data_type,), log, quiet)
        if self.has_checksum:
            checksum = Checksum.crc(data[0:12])
            if checksum != self.checksum and self.checksum:
                self._error('Inconsistent checksum (%04x/%04x)' % (checksum, self.checksum), log, quiet)

    def repair(self, data, log, header_size=None, protocol_version=None, profile_version=None):
        if header_size is not None and self.header_size != header_size:
            log.warning('Changing header size: %d -> %d' % (self.header_size, header_size))
            self.header_size = header_size
            self.data = [0] * self.header_size
            self.data[0] = self.header_size
        if protocol_version is not None and self.protocol_version != protocol_version:
            log.warning('Changing protocol version: %d -> %d' % (self.protocol_version, protocol_version))
            self.protocol_version = protocol_version
            self.data[1] = self.protocol_version
        if profile_version is not None and self.profile_version != profile_version:
            log.warning('Changing profile version: %d -> %d' % (self.profile_version, profile_version))
            self.profile_version = profile_version
            self.data[2:4] = pack('<H', self.profile_version)
        data_size = len(data) - len(self) - 2
        if data_size != self.data_size:
            log.warning('Fixing header data size: %d -> %d' % (self.data_size, data_size))
            self.data_size = data_size
            self.data[4:8] = pack('<I', self.data_size)
        if self.data_type != FIT:
            log.warning('Fixing header data type: %s -> %s' % (self.data_type, FIT))
            self.data_type = FIT
            self.data[8:12] = self.data_type
        if self.header_size >= 14:
            checksum = Checksum.crc(self.data[0:12])
            if not self.has_checksum or checksum != self.checksum:
                log.warning('Fixing header checksum: %04x -> %04x' %
                            (self.checksum if self.has_checksum else 0, checksum))
                self.checksum = checksum
                self.data[12:14] = pack('<H', checksum)
                self.has_checksum = True

    def describe_fields(self, types):
        yield '%s - header' % tohex(self.data[0:1])
        yield '%s - protocol version' % tohex(self.data[1:2])
        yield '%s - profile version' % tohex(self.data[2:4])
        yield '%s - data size' % tohex(self.data[4:8])
        yield '%s - data type' % tohex(self.data[8:12])
        if self.has_checksum:
            yield '%s - checksum' % tohex(self.data[12:14])


class Defined(Token):

    __slots__ = ('definition', 'timestamp')

    def __init__(self, tag, data, state, local_message_type):
        self.definition = state.definitions[local_message_type]
        if self.definition.timestamp_field:
            self.__parse_timestamp(data, state)
        self.timestamp = state.timestamp
        super().__init__(tag, True, data[0:self.definition.size])
        if self.definition.global_message_no == 206:
            self.__parse_field_definition(state)
            self.is_user = False

    def __parse_timestamp(self, data, state):
        field = self.definition.timestamp_field
        state.timestamp = field.field.type.parse(data[field.start:field.finish], 1,
                                                 self.definition.endian, state.timestamp)[0]

    def __parse_field_definition(self, state):
        record = self.parse().force()
        developer_index = record.attr.developer_data_index[0][0]
        number = record.attr.field_definition_number[0][0]
        # todo - we don't really need to convert name to type just to extract name below
        base_type = state.types.base_types[
            state.types.profile_to_type('fit_base_type').profile_to_internal(
                record.attr.fit_base_type_id[0][0])]
        # todo - more fields (optional)
        name = record.attr.field_name[0][0]
        units = record.attr.units[0][0]
        state.dev_fields[developer_index][number] = \
            TypedField(state.log, name, number, units, None, None, None, base_type.name, state.types)

    def parse(self, **options):
        return self.definition.message.parse(self.data, self.definition, self.timestamp, **options)

    def describe_fields(self, types):
        yield '%s - header (msg %d - %s)' % \
              (tohex(self.data[0:1]), self.data[0] & 0x0f, self.definition.message.name)
        for field in sorted(self.definition.fields, key=lambda field: field.start):
            if field.name == 'timestamp':
                yield '%s - %s (%s) %s' % (tohex(self.data[field.start:field.finish]), field.name,
                                           field.base_type.name, self.timestamp)
            else:
                yield '%s - %s (%s)' % (tohex(self.data[field.start:field.finish]), field.name, field.base_type.name)

    def describe_csv(self):
        record = self.parse(map_values=False, raw_time=True, rtn_composite=True)
        yield self.__class__.__name__
        yield self.definition.local_message_type
        yield record.name
        for name, (values, units) in record.data:
            yield name
            yield '' if values is None else '|'.join(str(value) for value in values)
            yield '' if units is None else units


class Data(Defined):

    __slots__ = ()

    def __init__(self, data, state):
        super().__init__('DTA', data, state, data[0] & 0x0f)


class CompressedTimestamp(Defined):

    __slots__ = ()

    def __init__(self, data, state):
        offset = data[0] & 0x1f
        if not state.timestamp:
            raise Exception('Compressed timestamp with no preceding absolute timestamp')
        if not isinstance(state.timestamp, dt.datetime):
            raise Exception('int timestamp')
        timestamp = time_to_timestamp(state.timestamp)
        rollover = offset < timestamp & 0x1f
        state.timestamp = timestamp_to_time((timestamp & 0xffffffe0) + offset + (0x20 if rollover else 0))
        super().__init__('TIM', data, state, data[0] & 0x60 >> 5)


class Field:

    def __init__(self, size, field, base_type):

        self.size = size
        self.field = field
        self.name = self.field.name if self.field else 'unknown'
        self.base_type = base_type

        self.count = self.size // base_type.size
        # set by definition later
        self.start = 0
        self.finish = 0


class Definition(Token):

    def __init__(self, data, state, overhead=6, tag='DFN'):
        self.local_message_type = data[0] & 0x0f
        self.is_user = False
        self.references = set()
        self.timestamp_field = None
        self.endian = data[2] & 0x01
        self.global_message_no = unpack('<>'[self.endian]+'H', data[3:5])[0]
        self.message = state.messages.number_to_message(self.global_message_no)
        self.identity = Identity(self.message.name, state.definition_counter)
        self.fields = self._process_fields(self._make_fields(data, state))
        super().__init__(tag, False, data[0:overhead+3*len(self.fields)])
        state.definitions[self.local_message_type] = self

    def _make_fields(self, data, state):
        yield from self.__fields(data, state.types)

    def __fields(self, data, types):
        for i in range(data[5]):
            yield self.__field(data[6 + i * 3:6 + (i + 1) * 3], self.message, types)

    def __field(self, data, message, types):
        number, size, base = data
        try:
            field = message.number_to_field(number)
        except KeyError:
            field = None
        base_type = types.base_types[base & 0xf]
        return Field(size, field, base_type)

    def _process_fields(self, fields):
        offset = 1  # header
        fields = tuple(fields)
        for field in fields:
            field.start = offset
            offset += field.size
            field.finish = offset
            if field.field and \
                    (field.field.number == TIMESTAMP_GLOBAL_TYPE or field.field.name == 'timestamp_16'):
                self.timestamp_field = field
            if field.field and isinstance(field.field, DynamicField):
                self.references.update(field.field.references)
        self.size = offset
        return tuple(sorted(fields, key=lambda field: 1 if field.field and isinstance(field.field, DynamicField) else 0))

    def accumulate(self, field, values):
        # todo - need test for this (currently clearly broken)
        n = len(values)
        if field in self.__sums:
            sum = self.__sums[field]
            while len(sum) < n:
                self.__sums += (0,)
            while len(values) < len(sum):
                values += (0,)
            sum = tuple(s + v for s, v in zip(sum, values))
            self.__sums[field] = sum
        else:
            self.__sums[field] = values
        return self.__sums[field][0:n]

    def describe_fields(self, types):
        yield '%s - header (msg %d)' % (tohex(self.data[0:1]), self.local_message_type)
        yield '%s - reserved' % tohex(self.data[1:2])
        yield '%s - architecture' % tohex(self.data[2:3])
        yield '%s - msg no (%s)' % (tohex(self.data[3:5]), self.message.name)
        yield '%s - no of fields' % tohex(self.data[5:6])
        for i in range(self.data[5]):
            data = self.data[6 + i*3:9 + i*3]
            field = self.__field(data, self.message, types)
            mult = '' if field.count == 1 else 'x%d' % field.count
            yield '  %s - fld %d: %s (%s%s)' % (tohex(data), i, field.name, field.base_type.name, mult)

    def describe_csv(self):
        yield 'Definition'
        yield self.local_message_type
        yield self.message.name
        for field in self.fields:
            yield field.name
            yield field.count
            yield ''


class DeveloperDefinition(Definition):

    def __init__(self, data, state):
        super().__init__(data, state, overhead=7, tag='DFX')

    def _make_fields(self, data, state):
        yield from super()._make_fields(data, state)
        for field_data in self.__field_data(data):
            yield self.__field(field_data, state.dev_fields)

    def __field_data(self, data):
        offset = data[5] * 3 + 7
        n_dev_fields = data[offset-1]
        for i in range(n_dev_fields):
            yield data[offset + i * 3:offset + (i + 1) * 3]

    def __field(self, data, dev_fields):
        number, size, developer_index = data
        field = dev_fields[developer_index][number]
        return Field(size, field, field.type)

    def describe_fields(self, types):
        yield from super().describe_fields(types)
        offset = self.data[5] * 3 + 7
        yield '%s - no of dev fields' % tohex(self.data[offset-1:offset])
        for field_data in self.__field_data(self.data):
            fdn, size, ddi = field_data
            yield '  %s - dev fld %d/%d' % (tohex(field_data), fdn, ddi)


class Checksum(ValidateToken):

    @staticmethod
    def crc(data):

        CRC = [0x0000, 0xCC01, 0xD801, 0x1400, 0xF001, 0x3C00, 0x2800, 0xE401,
               0xA001, 0x6C00, 0x7800, 0xB401, 0x5000, 0x9C01, 0x8801, 0x4400]

        checksum = 0
        for byte in data:
            tmp = CRC[checksum & 0xf]
            checksum = (checksum >> 4) & 0xfff
            checksum = checksum ^ tmp ^ CRC[byte & 0xf]
            tmp = CRC[checksum & 0xf]
            checksum = (checksum >> 4) & 0xfff
            checksum = checksum ^ tmp ^ CRC[(byte >> 4) & 0xf]
        return checksum

    def __init__(self, data):
        super().__init__('CRC', False, data)
        self.checksum = unpack('<H', self.data)[0]

    def validate(self, all_data, log, quiet=False):
        checksum = self.crc(all_data[:-2])
        if checksum != self.checksum:
            self._error('Bad checksum (%04x/%04x)' % (checksum, self.checksum), log, quiet)

    def repair(self, data, log):
        checksum = self.crc(data[:-2])
        if checksum != self.checksum:
            log.warning('Fixing final checksum: %04x -> %04x' % (self.checksum, checksum))
            self.checksum = checksum
            self.data = pack('<H', checksum)

    def describe_fields(self, types):
        yield '%s - checksum' % tohex(self.data)


def token_factory(data, state):
    header = data[0]
    if header & 0x80:
        return CompressedTimestamp(data, state)
    else:
        if header & 0x40:
            if header & 0x20:
                return DeveloperDefinition(data, state)
            else:
                return Definition(data, state)
        else:
            return Data(data, state)


class State:

    def __init__(self, log, types, messages):
        self.log = log
        self.types = types
        self.messages = messages
        self.dev_fields = defaultdict(lambda: WarnDict(log, 'No definition for developer field %s'))
        self.definitions = WarnDict(log, 'No definition for local message type %s')
        self.definition_counter = Counter()
        self.timestamp = None

    def copy(self):
        copy = State(self.log, self.types, self.messages)
        copy.dev_fields.update(self.dev_fields)
        copy.definitions.update(self.definitions)
        copy.definition_counter.update(self.definition_counter)
        copy.timestamp = self.timestamp
        return copy





ACTIVE_DISTANCE = 'Active Distance'
ACTIVE_TIME = 'Active Time'
ACTIVE_SPEED = 'Active Speed'
ACTIVITY = 'Activity'
CUMULATIVE_STEPS_START = 'Cumulative Steps Start'
CUMULATIVE_STEPS_FINISH = 'Cumulative Steps Finish'
DAILY_STEPS = 'Daily Steps'
DISTANCE = 'Distance'
FATIGUE = 'Fatigue'
FITNESS = 'Fitness'
FTHR = 'FTHR'
HEART_RATE = 'Heart Rate'
HR_IMPULSE = 'HR Impulse'
HR_ZONE = 'HR Zone'
LATITUDE = 'Latitude'
LONGITUDE = 'Longitude'
MAX_HR = 'MaxHR'  # no spaces because constant name
MAX_SPEED = 'MaxSpeed'  # no spaces because constant name
MEDIAN_KM_TIME = 'Median %dkm Time'
MEDIAN_KM_TIME_ANY = 'Median % Time'
PERCENT_IN_Z = 'Percent in Z%d'
PERCENT_IN_Z_ANY = 'Percent in Z%'
SEGMENT_TIME = 'Segment Time'
SEGMENT_DISTANCE = 'Segment Distance'
SEGMENT_HEART_RATE = 'Segment Heart Rate'
SPEED = 'Speed'
SPHERICAL_MERCATOR_X = 'Spherical Mercator X'
SPHERICAL_MERCATOR_Y = 'Spherical Mercator Y'
STEPS = 'Steps'
REST_HR = 'Rest HR'
TIME = 'Time'
TIME_IN_Z = 'Time in Z%d'
MAX_MED_HR_M = 'Max Med HR %dm'
MAX_MED_HR_M_ANY = 'Max Med HR %'

HR_MINUTES = (5, 10, 15, 20, 30, 60, 90, 120, 180)

MAX = '[max]'
MIN = '[min]'
CNT = '[cnt]'
AVG = '[avg]'
SUM = '[sum]'
MSR = '[msr]'   # measure (separate table, less efficient)


def summaries(*args):
    return ','.join(args)


M = 'm'
S = 's'
H = 'h'
DEG = 'deg'
KMH = 'km/h'
PC = '%'
BPM = 'bpm'
KG = 'kg'
STEPS_UNITS = 'steps'
ENTRIES = 'entries'


def round_km():
    yield from range(5, 21, 5)
    yield from range(25, 76, 25)
    yield from range(100, 251, 50)
    yield from range(300, 1001, 100)




import datetime as dt

import numpy as np
from bokeh.layouts import column, row

from ch2.stoats.display.nearby import nearby_any_time
from .data_frame import interpolate_to
from .plot import dot_map, line_diff, cumulative, health, activity, heart_rate_zones
from .server import SingleShotServer
from ..data import statistics
from ..data.data_frame import set_log, session, get_log
from ..lib.date import format_time, format_seconds, local_date_to_time
from ..squeal import ActivityGroup, ActivityJournal
from ..stoats.calculate.monitor import MonitorStatistics
from ..stoats.names import SPEED, DISTANCE, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y, TIME, FATIGUE, FITNESS, \
    REST_HR, DAILY_STEPS, ACTIVE_DISTANCE, ACTIVE_TIME, HR_ZONE

DISTANCE_KM = '%s / km' % DISTANCE
SPEED_KMH = '%s / kmh' % SPEED
HR_10 = 'HR Impulse / 10s'

LOG_FITNESS = 'Log %s' % FITNESS
LOG_FATIGUE = 'Log %s' % FATIGUE

TEMPLATE = '''
{% block css_resources %}
  {{ super() }}
  <style>
body {
  background-color: white;
}
.centre {
  text-align: center
}
.centre > div {
  display: inline-block;
}
table {
  margin: 20px;
  border-spacing: 10px;
}
  </style>
{% endblock %}
{% block inner_body %}
  <div class='centre'>
  <h1>{{ header }}</h1>
  {{ super() }}
  {{ caption }}
  </div>
{% endblock %}
'''


def comparison(log, s, aj1, aj2=None):

    set_log(log)
    names = [SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y, DISTANCE, SPEED, HR_ZONE, HR_10]

    st1 = statistics(s, *names, source_ids=[aj1.id])
    st1[DISTANCE_KM] = st1[DISTANCE]/1000
    st1[SPEED_KMH] = st1[SPEED] * 3.6
    st1_10 = interpolate_to(st1, HR_10)

    if aj2:
        st2 = statistics(s, *names, source_ids=[aj2.id])
        st2[DISTANCE_KM] = st2[DISTANCE]/1000
        st2[SPEED_KMH] = st2[SPEED] * 3.6
        st2_10 = interpolate_to(st2, HR_10)

    side = 300
    mx, mn = st1_10[HR_10].max(), st1_10[HR_10].min()
    st1_10['size'] = side * ((st1_10[HR_10] - mn) / (mx - mn)) ** 3 / 10
    x1, y1 = st1_10[SPHERICAL_MERCATOR_X], st1_10[SPHERICAL_MERCATOR_Y]
    if aj2:
        x2, y2 = st2_10[SPHERICAL_MERCATOR_X], st2_10[SPHERICAL_MERCATOR_Y]
    else:
        x2, y2 = None, None
    map = dot_map(side, x1, y1, st1_10['size'], x2, y2)

    y1 = st1_10[HR_10]
    y1.index = st1_10[DISTANCE_KM]  # could be left as time
    if aj2:
        y2 = st2_10[HR_10]
        y2.index = st2_10[DISTANCE_KM]
    else:
        y2 = None
    hr10_line = line_diff(700, 200, DISTANCE_KM, y1, y2)
    hr10_cumulative = cumulative(200, 200, y1, y2)

    hrz_histogram = heart_rate_zones(200, 200, st1_10[HR_ZONE])

    y1 = st1_10[DISTANCE_KM]
    c1 = st1_10[SPEED_KMH]
    if aj2:
        y2 = st2_10[DISTANCE_KM]
        c2 = st2_10[SPEED_KMH]
    else:
        c2 = None
    dist_line = line_diff(500, 200, TIME, y1, y2)
    dist_cumulative = cumulative(200, 200, c1, c2)

    finish = aj1.finish + dt.timedelta(days=1)  # to show new level
    start = finish - dt.timedelta(days=365)

    st_ff = statistics(s, FITNESS, FATIGUE, DAILY_STEPS, start=start, finish=finish)
    st_hr = statistics(s, REST_HR, owner=MonitorStatistics, start=start, finish=finish)
    st_ff[LOG_FITNESS] = np.log10(st_ff[FITNESS])
    st_ff[LOG_FATIGUE] = np.log10(st_ff[FATIGUE])

    health_line = health(600, 150, st_ff[LOG_FITNESS], st_ff[LOG_FATIGUE], st_hr[REST_HR])

    ajs = s.query(ActivityJournal). \
        join(ActivityGroup). \
        filter(ActivityJournal.start >= start,
               ActivityJournal.start < finish,
               ActivityGroup.id == aj1.activity_group_id
               ).all()
    st_ac = statistics(s, ACTIVE_TIME, ACTIVE_DISTANCE, source_ids=[aj.id for aj in ajs])

    activity_line = activity(600, 150, st_ff[DAILY_STEPS], st_ac[ACTIVE_TIME])  # last could be distance

    caption = '<div><table><tr><th>Name</th><th>Date</th><th>Duration</th><th>Distance</th></tr>'
    source_ids = [aj1.id]
    if aj2:
        source_ids.append(aj2.id)
    st = statistics(s, ACTIVE_TIME, ACTIVE_DISTANCE, source_ids=source_ids)
    st_aj = st.loc[st.index == aj1.start]
    caption += '<tr><td>%s</td><td>%s</td><td>%s</td><td>%.3f km</td></tr>' % \
               (aj1.name, format_time(aj1.start),
                format_seconds(st_aj[ACTIVE_TIME][0]), st_aj[ACTIVE_DISTANCE][0] / 1000)
    if aj2:
        st_aj = st.loc[st.index == aj2.start]
        caption += '<tr><td>%s</td><td>%s</td><td>%s</td><td>%.3f km</td></tr>' % \
                   (aj2.name, format_time(aj2.start),
                    format_seconds(st_aj[ACTIVE_TIME][0]), st_aj[ACTIVE_DISTANCE][0] / 1000)
    caption += '</table></div>'

    SingleShotServer(log, column(row(hr10_line, hr10_cumulative),
                                 row(dist_line, dist_cumulative, hrz_histogram),
                                 row(column(health_line, activity_line),
                                     map)),
                     template=TEMPLATE, title='choochoo',
                     template_vars={
                         'header': ('%s' % aj1.name) + ((' v %s' % aj2.name) if aj2 else ''),
                         'caption': caption
                     })


if __name__ == '__main__':
    s = session('-v 5')
    day = local_date_to_time('2018-02-14')
    aj1 = s.query(ActivityJournal).filter(ActivityJournal.start >= day,
                                          ActivityJournal.start < day + dt.timedelta(days=1)).one()
    aj2 = nearby_any_time(s, aj1)[0]
    comparison(get_log(), s, aj1, aj2)
    comparison(get_log(), s, aj1, aj2)




from sys import stdout

from .args import PATH, DROP, OUTPUT, SLICES, RAW, WARN, MIN_SYNC_CNT, MAX_RECORD_LEN, MAX_DROP_CNT, MAX_BACK_CNT, \
    MAX_FWD_LEN, DISCARD, FORCE, VALIDATE, no, ADD_HEADER, HEADER_SIZE, PROTOCOL_VERSION, PROFILE_VERSION, mm
from ..fit.fix import fix
from ..fit.profile.profile import read_fit


def fix_fit(args, log, db):
    '''
## fix-fit

    > ch2 fix-fit PATH -o PATH --drop

Try to fix a corrupted fit file.

By default, the length and checksum are updated.

If `--header` is specified then a new header is prepended at the start of the data.

If `--slices` is specified then the given slices are taken from the data and used to construct a new file.

If `--drop` is specified then the program tries to find appropriate slices by discarding data until all the
remaining data can be parsed.

The length and checksums are updated as appropriate.

### Examples

    > ch2 fix-fit FILE.FIT --slices 1000:

Will attempt to drop the first 1000 bytes from the given file.

    > ch2 fix-fit data/tests/personal/8CS90646.FIT --drop --discard

Will attempt to fix the given file (in the test data from git).

    > ch2 fix-fit FILE.FIT --add-header --header-size 14 --slices :14,28:

Will prepend a new 14 byte header, drop the old 14 byte header, and fix the header values.
    '''

    data = read_fit(log, args[PATH])
    log.info('Read %d bytes' % len(data))

    if not args[FORCE]:
        log.warning('Records are not evaluated (%s)' % no(FORCE))

    data = fix(log, data, add_header=args[ADD_HEADER], drop=args[DROP], slices=args[SLICES],
               warn=args[WARN], force=args[FORCE], validate=args[VALIDATE],
               header_size=args[HEADER_SIZE], protocol_version=args[PROTOCOL_VERSION], profile_version=args[PROFILE_VERSION],
               min_sync_cnt=args[MIN_SYNC_CNT], max_record_len=args[MAX_RECORD_LEN],
               max_drop_cnt=args[MAX_DROP_CNT], max_back_cnt=args[MAX_BACK_CNT], max_fwd_len=args[MAX_FWD_LEN])

    out_path = args[OUTPUT]
    if out_path:
        with open(out_path, 'wb') as out:
            out.write(data)
        log.info('Wrote data to %s' % out_path)
    elif args[DISCARD]:
        log.info('Discarded output')
    elif args[RAW]:
        log.info('Writing binary data to stdout')
        stdout.buffer.write(data)
    else:
        log.info('Writing hex data to stdout')
        stdout.write(data.hex())






from sqlalchemy import inspect
from urwid import connect_signal, disconnect_signal


class Binder:

    # this works in two ways:
    # 1 - a single row is updated.  key values must be supplied in defaults and composite
    #     keys are possible, but modifying the key values raises an exception
    # 2 - rows can be navigated.  a single primary key only.  when changed, the session
    #     is committed and a new value read from the database.

    def __init__(self, log, session, widget, table=None, multirow=False, defaults=None, instance=None):
        if defaults is None: defaults = {}
        self.__log = log
        self.__session = session
        self.__widget = widget
        self.__multirow = multirow
        self.__defaults = defaults
        self.__ignore_changes = False
        self.__table = table if table else type(instance)
        self.__primary_keys = tuple(map(lambda column: column.name, inspect(self.__table).primary_key))
        if self.__multirow and len(self.__primary_keys) > 1:
            raise Exception('Composite key not compatible with multirow')
        if instance:
            self.instance = instance
            self.__from_database = all(getattr(instance, key) is not None for key in self.__primary_keys)
        else:
            self.instance = None
            self.__from_database = False
            self.__read()
        self.__bind()

    def __bind(self,):
        self.__log.debug('Binding %s to %s' % (self.instance, self.__widget))
        save_ignore, self.__ignore_changes = self.__ignore_changes, True
        try:
            for column in inspect(self.__table).columns:
                name = column.name
                value = getattr(self.instance, name)
                if not name.startswith('_'):
                    try:
                        widget = getattr(self.__widget, name)
                        self.__log.debug('Setting %s=%s on %s' % (name, value, widget))
                        while widget:
                            if self._try_bind(column, value, widget):
                                break
                            widget = self._try_descend(column, value, widget)
                    except AttributeError as e:
                        self.__log.warning('Cannot find %s member of %s (%s): %s' %
                                        (name, self.__widget, dir(self.__widget), e))
        finally:
            self.__ignore_changes = save_ignore

    def _try_descend(self, column, value, widget):
        if hasattr(widget, 'base_widget') and widget != widget.base_widget:
            return widget.base_widget
        elif hasattr(widget, '_wrapped_widget') and widget != widget._wrapped_widget:
            return widget._wrapped_widget
        else:
            self.__log.error('Cannot set %s on %s (%s)' % (column.name, widget, dir(widget)))
            return None

    def _try_bind(self, column, value, widget):
        if hasattr(widget, 'state'):
            self.__bind_state(column, value, widget)
            return True
        elif hasattr(widget, 'set_edit_text'):
            self.__bind_edit(column, value, widget)
            return True
        return False

    def __with_default(self, column, value):
        if not column.nullable and value is None:
            if column.server_default:
                value = column.server_default.arg
                setattr(self.instance, column.name, value)
            else:
                raise Exception('Column %s is not nullable, but has no default' % column)
        return value

    def __bind_state(self, column, value, widget):
        widget.state = self.__with_default(column, value)
        self.__connect(widget, column.name)

    def __bind_edit(self, column, value, widget):
        widget.set_edit_text(self.__with_default(column, value))
        self.__connect(widget, column.name)

    def __connect(self, widget, name):
        # we don't use weak args because we want the binder to be around as long as the widget
        user_args = [name]
        self.__log.debug('Disconnecting %s' % widget)
        disconnect_signal(widget, 'change', self.__change_callback, user_args=user_args)
        self.__log.debug('Connecting %s' % widget)
        connect_signal(widget, 'change', self.__change_callback, user_args=user_args)

    def __change_callback(self, name, widget, value):
        self.__log.debug('Change %s=%s for %s (%s)' % (name, value, widget, self.instance))
        if name in self.__primary_keys:
            if self.__multirow:
                self.__session.commit()
                self.__defaults[name] = value
                self.__read()
                self.__bind()
            elif value != getattr(self.instance, name):
                raise Exception('Primary key (%s) modified, but not multirow')
        else:
            if not self.__from_database and not self.__ignore_changes:
                self.__log.debug('Adding %s to session' % self.instance)
                self.__session.add(self.instance)
            self.__log.debug('Setting %s=%s on %s' % (name, value, self.instance))
            setattr(self.instance, name, value)

    def __default_instance(self):
        return self.__table(**self.__defaults)

    def __read(self):
        self.instance = None
        if self.__defaults:
            self.__log.debug('Reading new %s' % self.__table)
            query = self.__session.query(self.__table)
            for (k, v) in self.__defaults.items():
                query = query.filter(getattr(self.__table, k) == v)
            self.instance = query.one_or_none()
        if self.instance:
            self.__from_database = True
        else:
            self.__from_database = False
            self.__log.debug("No database entry, so creating default from %s" % self.__defaults)
            self.instance = self.__default_instance()

    def reset(self):
        """
        Revert the widget to the values in the database (or defaults, if it is not from there).
        """
        if self.instance:
            self.__log.debug('Refreshing %s' % self.instance)
            if self.__from_database:
                self.__session.refresh(self.instance)
                self.__bind()
            else:
                self.__log.debug('Expunging %s' % self.instance)
                self.__session.expunge(self.instance)
                self.instance = self.__default_instance()
                self.__bind()

    def delete(self):
        """
        Remove from the database.
        """
        if self.instance:
            self.__log.debug('Deleting %s' % self.instance)
            self.__session.delete(self.instance)
            self.instance = None




from urwid import Text, Pile

from ..calculate.activity import ActivityStatistics
from ..names import PERCENT_IN_Z_ANY
from ...squeal.tables.statistic import StatisticJournal, StatisticName


def build_zones(s, ajournal, width):
    body = []
    percent_times = s.query(StatisticJournal).join(StatisticName). \
        filter(StatisticJournal.time == ajournal.start,
               StatisticName.name.like(PERCENT_IN_Z_ANY),
               StatisticName.owner == ActivityStatistics,
               StatisticName.constraint == ajournal.activity_group) \
        .order_by(StatisticName.name).all()
    for zone, percent_time in reversed(list(enumerate(percent_times, start=1))):
        text = ('%d:' + ' ' * (width - 6) + '%3d%%') % (zone, int(0.5 + percent_time.value))
        column = 100 / width
        left = int((percent_time.value + 0.5 * column) // column)
        text_left = text[0:left]
        text_right = text[left:]
        body.append(Text([('zone-%d' % zone, text_left), text_right]))
    return Pile(body)




import datetime as dt

from ...lib.date import to_date, format_date
from ...lib.schedule import DOW, Schedule
from ...squeal.tables.topic import Topic
from ...squeal.utils import ORMUtils


class Assert:

    def _assert(self, value, msg):
        if not value:
            raise Exception(msg)


class Week(Assert, ORMUtils):

    def __init__(self, name=None, description=None, start=None, days=None):
        self.__name = name
        self.__description = description
        self.__start = to_date(start)
        self.__days = dict((key.lower(), value) for key, value in days.items())
        self.__n_weeks = max(len(day) for day in self.__days.values())
        self.__validate()

    def __validate(self):
        self._assert(self.__name, 'No name')
        self._assert(self.__description, 'No description')
        self._assert(self.__start, 'No start')
        self._assert(self.__days, 'No days')
        self._assert(self.__n_weeks, 'No notes defined in days')
        for key, day in self.__days.items():
            self._assert(key in DOW, 'Bad day: %s' % key)
            self._assert(len(day) == 0 or self.__n_weeks == len(day),
                         'Day %s of unusual length (%d/%d)' % (key, self.__n_weeks, len(day)))

    def create(self, log, db, parent='Plan', sort=10):
        with db.session_context() as s:
            parent = self.__create_parent(log, s, parent, sort)
            self.__create_children(log, s, parent, sort)

    def __create_parent(self, log, s, root, sort):
        if self.__start.weekday():
            log.warning('The start day (%s) is not a Monday, so the days will be rotated appropriately',
                     DOW[self.__start.weekday()])
        root = self._get_or_create(s, Topic, name=root)
        schedule = Schedule('')
        schedule.start = self.__start
        schedule.finish = self.__start + dt.timedelta(days=7 * self.__n_weeks)
        parent = Topic(parent=root, schedule=schedule, sort=sort,
                       name=self.__name, description=self.__description)
        s.add(parent)
        # extend root to include parent
        root.schedule = Schedule.include(root.schedule, parent.schedule)
        return parent

    def __create_children(self, log, s, parent, sort):
        date = self.__start
        for day in DOW:
            if day in self.__days:
                self.__days[day].create(log, s, parent, sort, date, self.__n_weeks)
            date += dt.timedelta(days=1)


class Day(Assert):

    def __init__(self, name=None, notes=None):
        self.__name = name
        self.__notes = notes if notes else []
        self._assert(self.__name, 'No name')

    def __len__(self):
        return len(self.__notes)

    def create(self, log, s, parent, sort, date, n_weeks):
        dow = date.weekday()
        schedule = Schedule('%s/w[%s]' % (format_date(date), DOW[dow]))
        schedule.start = date
        schedule.finish = date + dt.timedelta(days=7 * n_weeks)
        child = Topic(parent=parent, schedule=schedule, name=self.__name, sort=sort)
        s.add(child)
        for week, note in enumerate(self.__notes):
            diary = Topic(schedule=str(date + dt.timedelta(days=7 * week)), parent=child,
                          description=note)
            s.add(diary)




from sqlalchemy import or_, desc, distinct
from sqlalchemy.orm import aliased
from sqlalchemy.sql.functions import min
from urwid import Pile, Text, Columns

from . import JournalDiary
from ...lib.date import to_time, time_to_local_date
from ...lib.utils import label
from ...squeal import ActivityJournal
from ...squeal import ActivitySimilarity
from ...uweird.tui.decorators import Indent
from ...uweird.tui.widgets import SquareButton, ColSpace, ColText


def fmt(time):
    return to_time(time).strftime('%Y-%m-%d')


class NearbyDiary(JournalDiary):

    def _display_schedule(self, s, f, date, schedule=None):
        yield from []

    def _journal_date(self, s, f, ajournal, date):
        for constraint in [c[0] for c in
                           s.query(distinct(ActivitySimilarity.constraint)).
                                   order_by(ActivitySimilarity.constraint).all()]:
            rows = []
            rows += self._any_time(s, f, ajournal, constraint)
            rows += self._earlier(s, f, ajournal, constraint)
            if rows:
                yield Pile([Text(constraint),
                            Indent(Pile(rows))])

    def _on_press(self, w, time):
        self._diary._change_date(time_to_local_date(time))

    def _buttons(self, f, title, data):
        if data:
            btns = [(len(fmt(d.start)) + 2,
                     f(SquareButton(fmt(d.start), on_press=self._on_press, user_data=d.start))) for d in data]
            yield Columns([ColText('%s: ' % title, label), *btns, ColSpace()])

    def _any_time(self, s, f, ajournal, constraint):
        yield from self._buttons(f, 'Any Time', nearby_any_time(s, ajournal, constraint=constraint))

    def _earlier(self, s, f, ajournal, constraint):
        yield from self._buttons(f, 'Earlier', nearby_earlier(s, ajournal, constraint=constraint))


def single_constraint(s, ajournal):
    return s.query(distinct(ActivitySimilarity.constraint)). \
        filter(or_(ActivitySimilarity.activity_journal_lo_id == ajournal.id,
                   ActivitySimilarity.activity_journal_hi_id == ajournal.id)).scalar()


def nearby_earlier(s, ajournal, constraint=None, threshold=0.05):
    if constraint is None:
        constraint = single_constraint(s, ajournal)
    ajlo = aliased(ActivityJournal)
    ajhi = aliased(ActivityJournal)
    return [asm.activity_journal_lo
            if asm.activity_journal_lo != ajournal
            else asm.activity_journal_hi
            for asm in s.query(ActivitySimilarity).
                join(ajhi, ActivitySimilarity.activity_journal_hi_id == ajhi.id).
                join(ajlo, ActivitySimilarity.activity_journal_lo_id == ajlo.id).
                filter(or_(ActivitySimilarity.activity_journal_hi_id == ajournal.id,
                           ActivitySimilarity.activity_journal_lo_id == ajournal.id),
                       ActivitySimilarity.constraint == constraint,
                       or_(ajhi.id == ajournal.id, ajhi.start < ajournal.start),
                       or_(ajhi.id == ajournal.id, ajlo.start < ajournal.start),
                       ActivitySimilarity.similarity > threshold).
                order_by(desc(min(ajlo.start, ajhi.start))).limit(6).all()]


def nearby_any_time(s, ajournal, constraint=None, threshold=0.05):
    if constraint is None:
        constraint = single_constraint(s, ajournal)
    return [asm.activity_journal_lo
            if asm.activity_journal_lo != ajournal
            else asm.activity_journal_hi
            for asm in s.query(ActivitySimilarity).
                filter(or_(ActivitySimilarity.activity_journal_hi_id == ajournal.id,
                           ActivitySimilarity.activity_journal_lo_id == ajournal.id),
                       ActivitySimilarity.constraint == constraint,
                       ActivitySimilarity.similarity > threshold).
                order_by(desc(ActivitySimilarity.similarity)).limit(6).all()]




from sqlalchemy.sql import func

from . import IntervalCalculator
from .summary import SummaryStatistics
from ..names import STEPS, REST_HR, HEART_RATE, DAILY_STEPS, BPM, STEPS_UNITS, summaries, SUM, AVG, CNT, MIN, MAX, MSR
from ..read.monitor import MonitorImporter
from ...lib.date import local_date_to_time
from ...squeal.database import add, StatisticJournal
from ...squeal.tables.source import Interval, NoStatistics
from ...squeal.tables.statistic import StatisticJournalInteger, StatisticName


# this is really just a daily summary - maybe it should be implemented as such?
# but it would be very inefficient for most stats.  should intervals be improved somehow?


QUARTER_DAY = 6 * 60 * 60


class MonitorStatistics(IntervalCalculator):

    def _on_init(self, *args, **kargs):
        kargs['schedule'] = 'd'
        super()._on_init(*args, **kargs)

    def _run_calculations(self, schedule):
        with self._db.session_context() as s:
            try:
                for start, finish in Interval.missing_dates(self._log, s, schedule, self, MonitorImporter):
                    self._log.info('Processing monitor data from %s to %s' % (start, finish))
                    self._add_stats(s, start, finish)
                    # stealth load so clean out summary manually
                    Interval.clean_dates(s, start, finish, owner=SummaryStatistics)
            except NoStatistics:
                self._log.info('No monitor data to process')

    def _add_stats(self, s, start, finish):
        start_time, finish_time = local_date_to_time(start), local_date_to_time(finish)
        interval = add(s, Interval(schedule='d', owner=self, start=start, finish=finish))
        midpt = start_time + 0.5 * (finish_time - start_time)
        m0 = s.query(func.avg(func.abs(StatisticJournalInteger.time - midpt))).join(StatisticName). \
            filter(StatisticName.name == HEART_RATE,
                   StatisticName.owner == MonitorImporter,
                   StatisticJournalInteger.time < finish_time,
                   StatisticJournalInteger.time >= start_time,
                   StatisticJournalInteger.value > 0).scalar()
        self._log.debug('M0: %s' % m0)
        if m0 and abs(m0 - QUARTER_DAY) < 0.25 * QUARTER_DAY:  # not evenly sampled
            rest_heart_rate = s.query(func.min(StatisticJournalInteger.value)).join(StatisticName). \
                filter(StatisticName.name == HEART_RATE,
                       StatisticName.owner == MonitorImporter,
                       StatisticJournalInteger.time < finish_time,
                       StatisticJournalInteger.time >= start_time,
                       StatisticJournalInteger.value > 0).scalar()
            self._add_integer_stat(s, interval, REST_HR, summaries(AVG, CNT, MIN, MSR), rest_heart_rate, BPM, start_time)
        else:
            self._log.info('Insufficient coverage for %s' % REST_HR)
        daily_steps = s.query(func.sum(StatisticJournalInteger.value)).join(StatisticName). \
            filter(StatisticName.name == STEPS,
                   StatisticName.owner == MonitorImporter,
                   StatisticJournalInteger.time < finish_time,
                   StatisticJournalInteger.time >= start_time).scalar()
        self._add_integer_stat(s, interval, DAILY_STEPS, summaries(SUM, AVG, CNT, MAX, MSR),
                               daily_steps, STEPS_UNITS, start_time)
        self._log.debug('Added data for %s' % interval)

    def _add_integer_stat(self, s, journal, name, summary, value, units, time):
        if value is not None:
            StatisticJournalInteger.add(self._log, s, name, units, summary, self, None, journal, value, time)




from abc import abstractmethod

from ch2.stoats import DbPipeline
from ...lib.io import for_modified_files
from ...squeal.database import add
from ...squeal.tables.statistic import StatisticJournal, StatisticName


class AbortImport(Exception):
    pass


class AbortImportButMarkScanned(AbortImport):
    pass


class Importer(DbPipeline):

    def _on_init(self, *args, **kargs):
        super()._on_init(*args, **kargs)
        self.__statistics_cache = {}

    def _first(self, path, records, *names):
        try:
            return next(iter(record for record in records if record.name in names))
        except StopIteration:
            self._log.debug('No %s entry(s) in %s' % (str(names), path))
            raise AbortImportButMarkScanned()

    def _last(self, path, records, *names):
        save = None
        for record in records:
            if record.name in names:
                save = record
        if not save:
            self._log.debug('No %s entry(s) in %s' % (str(names), path))
            raise AbortImportButMarkScanned()
        return save

    def _run(self, paths, force=False):
        with self._db.session_context() as s:
            for_modified_files(self._log, s, paths, self._callback(), self, force=force)

    def _callback(self):
        def callback(file):
            self._log.debug('Scanning %s' % file)
            with self._db.session_context() as s:
                try:
                    self._import(s, file)
                    return True
                except AbortImport as e:
                    self._log.debug('Aborted %s' % file)
                    return isinstance(e, AbortImportButMarkScanned)
        return callback

    @abstractmethod
    def _import(self, s, path):
        pass

    def _create(self, s, name, units, summary, constraint, source, value, time, type):
        # cache statistic_name instances for speed (avoid flush on each query)
        key = (name, constraint)
        if key not in self.__statistics_cache:
            self.__statistics_cache[key] = \
                StatisticName.add_if_missing(self._log, s, name, units, summary, self, constraint)
        if key not in self.__statistics_cache or not self.__statistics_cache[key]:
            raise Exception('Failed to get StatisticName for %s' % key)
        return type(statistic_name=self.__statistics_cache[key], source=source, value=value, time=time)

    def _add(self, s, name, units, summary, constraint, source, value, time, type):
        return add(s, self._create(s, name, units, summary, constraint, source, value, time, type))





class NoMaximum(Exception): pass

# http://acooke.org/cute/EfficientS2.html
def expand_max(log, lo, hi, n, f):

    data = [(x, f(x)) for x in (lo + i * (hi - lo) / 4 for i in range(5))]
    x_max, fx_max = None, None

    for (x, fx) in data:
        if x_max is None or fx > fx_max:
            x_max, fx_max = x, fx

    try:
        for _ in range(n):
            log.info('%s' % data)
            while len(data) > 3:
                w = sum(x*fx for (i, (x, fx)) in enumerate(data)) / sum(fx for (x, fx) in data)
                m = sum(x for (x, fx) in data) / len(data)
                if w > m:
                    del data[0]
                else:
                    del data[-1]
            x = (data[0][0] + data[1][0]) / 2
            fx = f(x)
            if fx > fx_max:
                x_max = x
            data.insert(1, (x, fx))
            x = (data[-2][0] + data[-1][0]) / 2
            fx = f(x)
            if fx > fx_max:
                x_max = x
            data.insert(-1, (x, fx))
    except ZeroDivisionError:
        pass

    return x_max, fx_max




import datetime as dt

import pandas as pd
from bokeh.models import NumeralTickFormatter, PrintfTickFormatter, Range1d, LinearAxis
from bokeh import palettes, tile_providers
from bokeh.plotting import figure

from .data_frame import interpolate_to_index, delta_patches, closed_patch
from ..stoats.names import TIME, HR_ZONE


def dot_map(n, x1, y1, size, x2=None, y2=None):

    f = figure(plot_width=n, plot_height=n, x_axis_type='mercator', y_axis_type='mercator')
    f.circle(x=x1, y=y1, line_alpha=0, fill_color='red', size=size, fill_alpha=0.03)
    f.line(x=x1, y=y1, line_color='black')

    if x2 is not None:
        f.line(x=x2, y=y2, line_color='grey')

    f.add_tile(tile_providers.STAMEN_TERRAIN, alpha=0.1)
    f.axis.visible = False

    f.toolbar_location = None
    return f


def line_diff(nx, ny, xlabel, y1, y2=None):

    is_time = isinstance(y1.index[0], dt.datetime)
    f = figure(plot_width=nx, plot_height=ny, x_axis_type='datetime' if is_time else 'linear')

    if is_time:
        f.xaxis.axis_label = 'Time'
        f.xaxis[0].formatter = NumeralTickFormatter(format='00:00:00')
        y1.index = (y1.index - y1.index[0]).total_seconds()
    else:
        f.xaxis.axis_label = xlabel
        f.xaxis[0].formatter = PrintfTickFormatter(format='%.3f')

    scale = y1.max()
    if y2 is not None:
        scale = max(scale, y2.max())

    f.line(x=y1.index, y=y1, color='black')
    f.y_range = Range1d(start=0, end=scale * 1.1)
    f.yaxis.axis_label = y1.name

    if y2 is not None:
        if is_time:
            y2.index = (y2.index - y2.index[0]).total_seconds()
        y2 = interpolate_to_index(y1, y2)
        f.line(x=y2.index, y=y2, color='grey')

        y1, y2, range = delta_patches(y1, y2)
        f.extra_y_ranges = {'delta': range}
        if y1 is not None:
            f.patch(x=y1.index, y=y1, color='green', alpha=0.1, y_range_name='delta')
        if y2 is not None:
            f.patch(x=y2.index, y=y2, color='red', alpha=0.1, y_range_name='delta')

    f.toolbar_location = None
    return f


def cumulative(nx, ny, y1, y2=None, sample=10):

    y1 = y1.sort_values(ascending=False).reset_index(drop=True)
    scale = y1.max()
    if y2 is not None:
        y2 = y2.sort_values(ascending=False).reset_index(drop=True)
        scale = max(scale, y2.max())

    f = figure(plot_width=nx, plot_height=ny,
               x_range=Range1d(start=y1.index.max() * sample, end=0),
               x_axis_type='datetime',
               x_axis_label=TIME,
               y_range=Range1d(start=0, end=scale * 1.1),
               y_axis_location='right',
               y_axis_label=y1.name)
    f.xaxis[0].formatter = NumeralTickFormatter(format='00:00:00')

    f.line(x=y1.index * sample, y=y1, color='black')
    if y2 is not None:
        f.line(x=y2.index * sample, y=y2, color='grey')
        y1, y2, range = delta_patches(y1, y2)
        f.extra_y_ranges = {'delta': range}
        f.patch(x=y1.index * sample, y=y1, color='green', alpha=0.1, y_range_name='delta')
        f.patch(x=y2.index * sample, y=y2, color='red', alpha=0.1, y_range_name='delta')

    f.toolbar_location = None
    return f


def health(nx, ny, ftn, ftg, hr):

    f = figure(plot_width=nx, plot_height=ny, x_axis_type='datetime')
    f.xaxis.axis_label = 'Date'

    max_f = ftn.max() * 1.1
    min_f = ftn.min() * 0.9

    f.y_range = Range1d(start=min_f, end=max_f)
    f.yaxis.axis_label = '%s, %s' % (ftn.name, ftg.name)
    f.yaxis[0].formatter = PrintfTickFormatter(format='')

    patch = closed_patch(ftg, zero=min_f)
    f.patch(x=patch.index, y=patch, color='grey', alpha=0.2)
    f.line(x=ftn.index, y=ftn, color='black')

    hr = hr.dropna()
    if len(hr):
        f.extra_y_ranges = {hr.name: Range1d(start=25, end=70)}
        f.add_layout(LinearAxis(y_range_name=hr.name, axis_label=hr.name), 'right')
        f.circle(x=hr.index, y=hr, color='red', alpha=0.2, y_range_name=hr.name)

    f.toolbar_location = None
    return f


def activity(nx, ny, st, at):

    f = figure(plot_width=nx, plot_height=ny, x_axis_type='datetime')
    f.xaxis.axis_label = 'Date'

    st = st.dropna()
    if len(st):
        f.y_range = Range1d(start=0, end=1.1 * st.max())
        f.yaxis.axis_label = st.name
        f.vbar(x=st.index, width=dt.timedelta(hours=20), top=st, fill_color='grey', fill_alpha=0.4, line_alpha=0)
    else:
        f.yaxis[0].visible = False

    at = at.dropna()
    if len(at):
        f.extra_y_ranges = {at.name: Range1d(start=0, end=at.max() * 1.1)}
        f.add_layout(LinearAxis(y_range_name=at.name, axis_label=at.name), 'right')
        f.circle(x=at.index, y=at, color='black', fill_alpha=0, y_range_name=at.name)
        f.yaxis[1].formatter = PrintfTickFormatter(format='')

    f.toolbar_location = None
    return f


def heart_rate_zones(nx, ny, hrz):

    max_z, n_sub = 5, 5

    bins = pd.interval_range(start=1, end=max_z+1, periods=n_sub * max_z, closed='left')
    c = [palettes.Inferno[6][int(b.left)-1] for b in bins]
    hrz_categorized = pd.cut(hrz, bins)
    counts = hrz_categorized.groupby(hrz_categorized).count()

    f = figure(plot_width=nx, plot_height=ny, x_range=Range1d(start=1, end=max_z+1), x_axis_label=HR_ZONE)
    f.quad(left=counts.index.categories.left, right=counts.index.categories.right, top=counts, bottom=0,
           color=c, fill_alpha=0.2)
    f.yaxis.visible = False

    f.toolbar_location = None
    return f

